from enum import Enum
from io import StringIO
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Literal

import yaml
from loguru import logger
from pydantic import (
    BaseModel,
    DirectoryPath,
    Field,
    field_validator,
    ConfigDict,
    ValidationInfo,
)
from uuid import UUID, uuid4

# from pydantic.typing import Literal  # type: ignore

from llmsearch.models.config import (
    HuggingFaceModelConfig,
    LlamaModelConfig,
    OpenAIModelConfig,
    AzureOpenAIModelConfig,
)

models_config = {
    "llamacpp": LlamaModelConfig,
    "openai": OpenAIModelConfig,
    # "auto-gptq": AutoGPTQModelConfig,
    "huggingface": HuggingFaceModelConfig,
    "azureopenai": AzureOpenAIModelConfig,
}


def create_uuid() -> str:
    return str(uuid4())


class Document(BaseModel):
    """Interface for interacting with a document."""

    page_content: str
    metadata: dict = Field(default_factory=dict)


class CustomDocumentExtension(str, Enum):
    md = "md"
    pdf = "pdf"
    docx = "docx"
    doc = "doc"


class RerankerModel(Enum):
    MARCO_RERANKER = "marco"
    BGE_RERANKER = "bge"


class EmbeddingModelType(str, Enum):
    huggingface = "huggingface"
    instruct = "instruct"
    sentence_transformer = "sentence_transformer"


class EmbeddingModel(BaseModel):
    model_config = ConfigDict()
    model_config["protected_namespaces"] = ()

    type: EmbeddingModelType
    model_name: str
    additional_kwargs: dict = Field(default_factory=dict)


class DocumentPathSettings(BaseModel):
    doc_path: Union[DirectoryPath, str]
    """Defines document folder for a given document set."""

    exclude_paths: List[Union[DirectoryPath, str]] = Field(default_factory=list)
    """List of folders to exclude from scanning."""

    scan_extensions: List[str]
    """List of extensions to scan."""

    additional_parser_settings: Dict[str, Any] = Field(default_factory=dict)
    """Optional parser settings (parser dependent)"""

    passage_prefix: str = ""


    label: str = ""  # Optional label, will be included in the metadata
    """Optional label for the document set, will be included in the metadata."""

    @field_validator("additional_parser_settings")
    def validate_extension(cls, value):
        for ext in value.keys():
            if ext not in CustomDocumentExtension.__members__:
                raise TypeError(
                    f"Custom parser settings aren't supported for document extension {value}. Supported: {CustomDocumentExtension.__members__}"
                )
        return value


class EmbedddingsSpladeConfig(BaseModel):
    n_batch: int = 3


class EmbeddingsConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")

    embedding_model: EmbeddingModel = EmbeddingModel(
        type=EmbeddingModelType.instruct, model_name="hkunlp/instructor-large"
    )
    """Specifies embedding model to use for dense embeddings."""

    embeddings_path: Union[DirectoryPath, str]
    """Specifies output folder for embeddings."""

    document_settings: List[DocumentPathSettings]
    """Defines settings for one or more document sets."""

    chunk_sizes: List[int] = [1024]
    """List of chunk sizes for text chunking, supports multiples sizes."""

    splade_config: EmbedddingsSpladeConfig = EmbedddingsSpladeConfig(n_batch=5)
    """Specifies settings for sparse embeddings (SPLADE)."""

    @property
    def labels(self) -> List[str]:
        """Returns list of labels in document settings"""
        return [setting.label for setting in self.document_settings if setting.label]


class ObsidianAdvancedURI(BaseModel):
    append_heading_template: str


class SuffixAppend(BaseModel):
    append_template: str


class ReplaceOutputPath(BaseModel):
    substring_search: str
    substring_replace: str


class HydeSettings(BaseModel):
    enabled: bool = False
    hyde_prompt: str = "Write a short passage to answer the question: {question}"


class RerankerSettings(BaseModel):
    enabled: bool = True
    model: RerankerModel = RerankerModel.BGE_RERANKER


class MultiQuerySettings(BaseModel):
    enabled: bool = False
    multiquery_prompt: str = """You are a helpful assistant that generates multiple questions based on the source question.
    Generate {n_versions} additional related questions related to: ```{question}```.
    
    Suggest only short questions without compound sentences. Suggest a variety of questions that cover different aspects of the topic.
    Make sure they are complete questions, and that they are related to the original question.

    Generated questions should be separated by newlines, but shouldn't be enumerated.
    """
    # multiquery_prompt: str =  """You are an AI language model assistant. Your task is
    # to generate {n_versions} different versions of the given user
    # question. The questions can be generated using domain specific language to clarify the intent. Provide these alternative
    # questions separated by newlines. Don't enumerate the alternative questions. Original question: {question}"""
    n_versions: int = 5


class SemanticSearchConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    search_type: Literal["mmr", "similarity"]
    """Configure search type, currently only similarity can be used."""

    replace_output_path: List[ReplaceOutputPath] = Field(default_factory=list)

    obsidian_advanced_uri: Optional[ObsidianAdvancedURI] = None
    
    append_suffix: Optional[SuffixAppend] = None
    """Allows to append suffix to document URL. Useful for deep linking to allow opening with external application, e.g. Obsidian."""
    
    reranker: RerankerSettings = RerankerSettings()
    """Configures re-ranker settings."""
    
    max_k: int = 15
    """Maximum number of documents to retrieve for dense OR sparse embedding (if using both, number of documents will be k*2)"""
    
    max_char_size: int = 2048
    """Maximum character size for query + documents to fit into context window of LLM."""
    
    query_prefix: str = ""
    """Prefix query with string BEFORE retrieval using embedding model."""
    
    hyde: HydeSettings = HydeSettings()
    """Optional configuration for HyDE."""
    
    multiquery: MultiQuerySettings = MultiQuerySettings()
    """Optional configuration for Multi-query."""


class LLMConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")
    model_config["protected_namespaces"] = ()

    type: str
    params: dict

    @field_validator("params")
    def validate_params(cls, value, info: ValidationInfo):
        values = info.data
        type_ = values.get("type")
        if type_ not in models_config:
            raise TypeError(f"Uknown model type {value}. Allowed types: ")

        config_type = models_config[type_]
        logger.info(
            f"Loading model paramaters in configuration class {config_type.__name__}"
        )
        config = config_type(
            **value
        )  # An attempt to force conversion to the required model config
        return config


class SemanticSearchOutput(BaseModel):
    chunk_link: str
    chunk_text: str
    metadata: dict


class ResponseModel(BaseModel):
    id: UUID = Field(default_factory=create_uuid)
    question: str
    response: str
    average_score: float
    semantic_search: List[SemanticSearchOutput] = Field(default_factory=list)
    hyde_response: str = ""


class Config(BaseModel):
    cache_folder: Path
    """Configures path to cache LLM and embedding models."""

    embeddings: EmbeddingsConfig
    """Configures document paths and embedding settings."""

    semantic_search: SemanticSearchConfig
    """Confgures semantic search settings."""

    llm: Optional[LLMConfig] = None
    "Don't use directly."

    persist_response_db_path: Optional[str] = None
    """Optional path for SQLite database for results storage."""

    def check_embeddings_exist(self) -> bool:
        """Checks if embedings exist in the specified folder"""

        p_splade = (
            Path(self.embeddings.embeddings_path) / "splade" / "splade_embeddings.npz"
        )
        p_embeddings = Path(self.embeddings.embeddings_path)
        all_parquets = list(p_embeddings.glob("*.parquet"))
        return p_splade.exists() and len(all_parquets) > 0


def get_config(path: Union[str, Path]) -> Config:
    with open(path, "r") as f:
        conf_dict = yaml.safe_load(f)
    return Config(**conf_dict)


def get_doc_with_model_config(doc_config, model_config) -> Config:
    """Loads doc and model configurations, combines, and returns an instance of Config"""

    doc_config_dict = load_yaml_file(doc_config)
    model_config_dict = load_yaml_file(model_config)

    config_dict = {**doc_config_dict, **model_config_dict}
    return Config(**config_dict)


def load_yaml_file(config) -> dict:
    if isinstance(config, str):
        logger.info(f"Loading doc config from a file: {config}")
        with open(config, "r") as f:
            string_data = f.read()
    else:
        stringio = StringIO(config.getvalue().decode("utf-8"))
        string_data = stringio.read()

    config_dict = yaml.safe_load(string_data)
    return config_dict



# import os

# from llmsearch.chroma import ChromaVS
# from llmsearch.llm import HuggingFaceWrapper, LLMOpenAIWrapper
# from llmsearch.parsers.markdown_v1 import markdown_parser
# from termcolor import colored, cprint

# ## Define fodlers

# STORAGE_FOLDER_ROOT = "/storage/llm/"
# CACHE_FOLDER_ROOT = os.path.join(STORAGE_FOLDER_ROOT, "cache")
# INDEX_PERSIST_FOLDER = os.path.join(STORAGE_FOLDER_ROOT, "index")
# DOC_FOLDER = os.path.join(STORAGE_FOLDER_ROOT, "docs")

# # Define embedding and LLM models
# # EMBEDDING_MODEL = "all-distilroberta-v1" # Assumes HuggingFace embedding.
# EMBEDDING_MODEL = "all-MiniLM-L6-v2" # Assumes HuggingFace embedding.
# URL_PREFIX = "obsidian://open?vault=knowledge-base&file="

# os.environ["SENTENCE_TRANSFORMERS_HOME"] = CACHE_FOLDER_ROOT
# os.environ['TRANSFORMERS_CACHE'] = os.path.join(CACHE_FOLDER_ROOT,  "transformers")
# os.environ['HF_HOME'] = os.path.join(CACHE_FOLDER_ROOT,  "hf_home")


# if __name__ == "__main__":
#     persist_folder = INDEX_PERSIST_FOLDER

#     model = LLMOpenAIWrapper(
#         embedding_model_name=EMBEDDING_MODEL, llm_model_name="gpt-3.5-turbo",
#         temperature=0,
#         max_input_size = 4096,
#         num_output = 2048
#     )

#     # model = HuggingFaceWrapper(
#     #     embedding_model_name=EMBEDDING_MODEL,
#     #     llm_model_name="databricks/dolly-v2-3b",
#     #     cache_folder=CACHE_FOLDER_ROOT,
#     #     max_input_size = 2048,
#     #     num_output = 1024
#     # )

#     vectordb = ChromaVS(
#         persist_folder=persist_folder,
#         collection_name="md_sample",
#         embedding_model_name=EMBEDDING_MODEL,
#         service_context=model.service_context,
#     )

#     # vectordb.create_index_from_folder(folder_path=DOC_FOLDER, extension="md", parser_func=markdown_parser)

#     index = vectordb.load_index()
#     # #  index = vectordb.create_index_from_folder(folder_path=DOC_FOLDER, extension="md", parser_func=markdown_parser, service_context = service_context)

#     # print(index)

#     query_engine = index.as_query_engine(mode="embedding", similarity_top_k=3,
#                                          response_mode="compact",
#                                          verbose=True)
#  r = query_engine.query("How to overwrite a table in Spark?")
# r = query_engine.query("How to convert from string to timestamp in Spark?")
# #r = query_engine.query("What is a difference between transformation and action in Spark?")
# r = query_engine.query("What is buffer pool manager?")

# r = query_engine.query("How to copy and paste in vim, in linux?")

# r = query_engine.query("How to connect to databricks using service principal?")

# r = query_engine.query("Provide minimum working skeleton for pyproject.toml")

# r = query_engine.query("What type of data should be contained in bronze table?")

#    r = query_engine.query("What is a difference between dataclass and regular classes in Python?")

# r = query_engine.query("Provide a code example to update delta table")

# r = query_engine.query("How to check if file exist in bash script?")

# r = query_engine.query("What is a kalman filter?")

# r = query_engine.query("What is a difference between data and non-data descriptor in Python?")

# r = query_engine.query("How to make a class hashable in Python?")

# r = query_engine.query("How data is stored on disk for analytical databases?")

# r = query_engine.query("What type of hashing schemes exist in databases?")

# r = query_engine.query("How to read YAML file into Pydantic model? Provide a code example")

# r = query_engine.query("Provide different methods to update dictionary in Python")

# r = query_engine.query("Provide a code example on how pass parameters to constructor of parent class in Python?")

# #r = query_engine.query("How to use sed to do replacement in strings and write it back to file? Provide an example command.")

# #r = query_engine.query("How to to recursively search files in sub-directories using pathlib in Python?")

# #r = query_engine.query("What is a difference between howto and tutorial in documentation?")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath('../src'))

project = 'LLM Search'
copyright = '2024, Denis L.'
author = 'Denis L.'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.duration", "sphinx.ext.autodoc", "sphinxcontrib.autodoc_pydantic" ]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']

autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_show_json = False
autodoc_pydantic_model_show_validator_members = False
autodoc_pydantic_model_show_validator_summary = False
autodoc_pydantic_field_list_validators = False


import os
from typing import Any, List

import langchain
import uvicorn
from dotenv import load_dotenv
from fastapi import Depends, FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from loguru import logger
from sqlalchemy.orm import Session

import llmsearch.database.crud as crud
from llmsearch.config import Config, get_config, ResponseModel
from llmsearch.database.config import get_local_session
from llmsearch.process import get_and_parse_response
from llmsearch.ranking import get_relevant_documents
from llmsearch.utils import LLMBundle, get_llm_bundle

load_dotenv()
langchain.debug = True

# Load the configuration
def read_config() -> Config:
    config_file = os.environ["FASTAPI_LLM_CONFIG"]

    if not config_file:
        raise SystemError(
            "Set 'FASTAPI_LLM_CONFG' environment variable to point to a model config file."
        )

    logger.info(f"Loading configuration from {config_file}")
    config = get_config(config_file)

    return config


# Dependency: Used to get the database in our endpoints.
def get_db() -> Session:
    """Creates a database session and makes sure to close it properly."""

    if config.persist_response_db_path is None:
        raise Exception(
            "Specify database filename in `persist_response_db_path` setting in config."
        )
    db_settings = get_local_session(db_path=config.persist_response_db_path)

    db = db_settings.SessionLocal()
    try:
        yield db
    finally:
        logger.info("Closing session...")
        db.close()


def load_llm() -> LLMBundle:
    """Loads a chain to use with the api"""

    logger.info("Loading LLM...")
    bundle = get_llm_bundle(config)
    return bundle


config = read_config()
llm_bundle = load_llm()

app = FastAPI()

# Enable CORS
origins = ["*"]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/")
def test():
    return {"message": "Welcome to LLMSearch API"}


@app.get("/llm", response_model=ResponseModel)
async def llmsearch(
    question: str, label: str = "", db: Session = Depends(get_db)
) -> Any:  # switch to async to block execution
    if label and (label not in config.embeddings.labels):
        raise HTTPException(
            status_code=404,
            detail=f"Label '{label}' doesn't exist. Use GET /labels to get a list of labels.",
        )

    output = get_and_parse_response(
        query=question,
        llm_bundle=llm_bundle,
        config=config,
        persist_db_session=db,
        label=label,
    )
    return output.dict()


@app.get("/semantic")
async def semanticsearch(question: str):
    docs = get_relevant_documents(
        original_query=question, queries = [question], llm_bundle=llm_bundle, config=config.semantic_search, label=""
    )
    return {"sources": docs}


@app.get("/labels")
async def labels() -> List[str]:
    return config.embeddings.labels


@app.post("/feedback")
async def llmfeedback(
    response_id: str,
    is_positive: bool,
    feedback_text: str = "",
    db: Session = Depends(get_db),
):
    try:
        crud.create_feedback(
            response_id=response_id,
            session=db,
            is_positive=is_positive,
            feedback_text=feedback_text,
        )
    except crud.ResponseInteractionLookupError:
        return HTTPException(
            status_code=404,
            detail=f"Reponse interaction with id '{response_id}' doesn't exist",
        )

    return {"status": "Update complete"}


def main():
    uvicorn.run(app, host="0.0.0.0", port=8000)


if __name__ == "__main__":
    main()



import torch
from transformers import AutoModelForSequenceClassification, AutoTokenizer
from typing import List
from llmsearch.config import SemanticSearchConfig

# from llmsearch.utils import LLMBundle
from typing import Tuple
import statistics

from loguru import logger
from sentence_transformers.cross_encoder import CrossEncoder

from llmsearch.config import Document


class MarcoReranker:
    def __init__(
        self, cross_encoder_model: str = "cross-encoder/ms-marco-MiniLM-L-6-v2"
    ) -> None:
        logger.info("Initializing Reranker...")
        self._crdss_encoder_model_name = cross_encoder_model
        self.model = CrossEncoder(cross_encoder_model)
        logger.info("Initialized MS-MARCO Reranker")

    def get_scores(self, query: str, docs: List[Document]) -> List[float]:
        logger.info("Reranking documents ... ")

        features = [[query, doc.page_content] for doc in docs]
        scores = self.model.predict(features).tolist()
        return scores


class BGEReranker:
    def __init__(self) -> None:
        self.tokenizer = AutoTokenizer.from_pretrained("BAAI/bge-reranker-base")
        self.model = AutoModelForSequenceClassification.from_pretrained(
            "BAAI/bge-reranker-base"
        )
        self.model.eval()
        logger.info("Initialized BGE-base Reranker")

    def get_scores(self, query: str, docs: List[Document]) -> List[float]:
        logger.info("Reranking documents ... ")
        features = [[query, doc.page_content] for doc in docs]
        with torch.no_grad():
            inputs = self.tokenizer(
                features,
                padding=True,
                truncation=True,
                return_tensors="pt",
                max_length=512,
            )
            scores = (
                self.model(**inputs, return_dict=True)
                .logits.view(
                    -1,
                )
                .float()
                .tolist()
            )
        return scores


def rerank(
    rerank_model, query: str, docs: List[Document]
) -> Tuple[float, List[Document]]:
    logger.info("Reranking documents ... ")
    scores = rerank_model.get_scores(query, docs)
    print(scores)
    for score, d in zip(scores, docs):
        d.metadata["score"] = score

    sorted_scores = sorted(scores, reverse=True)

    logger.info(sorted_scores)
    median_ = statistics.mean(sorted_scores[:5])
    return median_, [
        doc for doc in sorted(docs, key=lambda d: d.metadata["score"], reverse=True)
    ]


def get_relevant_documents(
    original_query: str,
    queries: List[str],
    llm_bundle,
    config: SemanticSearchConfig,
    label: str,
) -> Tuple[List[str], float]:
    most_relevant_docs = []
    docs = []

    # original_query = queries[0]
    sparse_retriever = llm_bundle.sparse_search

    # Get max_k  documents using sparse search

    current_reranker_score, reranker_score = -1e5, -1e5

    # Iterate over all available chunk sizes
    for chunk_size in llm_bundle.chunk_sizes:
        all_relevant_docs = []
        all_relevant_doc_ids = set()
        for query in queries:
            logger.debug("Evaluating query: {}", query)
            if config.query_prefix:
                logger.info(f"Adding query prefix for retrieval: {config.query_prefix}")
                query = config.query_prefix + query
            sparse_search_docs_ids, sparse_scores = sparse_retriever.query(
                search=query, n=config.max_k, label=label, chunk_size=chunk_size
            )

            logger.info(f"Stage 1: Got {len(sparse_search_docs_ids)} documents.")

            # Set a filter for current chunk size or skip filter if only one chunk size is present (considerably faster)
            filter = (
                {"chunk_size": chunk_size}
                if len(llm_bundle.chunk_sizes) > 1
                else dict()
            )

            # Add label to filter, if present
            if label:
                filter.update({"label": label})

            if (
                not filter
            ):  # if filter is empty (doesn't contain chunk_size or label), set it to None to speed up.
                filter = None

            logger.info(f"Dense embeddings filter: {filter}")

            res = llm_bundle.store.similarity_search_with_relevance_scores(
                query, k=config.max_k, filter=filter
            )
            dense_search_doc_ids = [r[0].metadata["document_id"] for r in res]

            # Create union of documents fetched using sprase and dense embeddings
            all_doc_ids = (
                set(sparse_search_docs_ids).union(set(dense_search_doc_ids))
            ).difference(all_relevant_doc_ids)
            logger.debug("NUMBER OF NEW DOCS to RETRIEVE: {}", len(all_doc_ids))
            if all_doc_ids:
                relevant_docs = llm_bundle.store.get_documents_by_id(
                    document_ids=list(all_doc_ids)
                )
                all_relevant_docs += relevant_docs
                all_relevant_doc_ids = all_relevant_doc_ids.union(all_doc_ids)

        # Choose chunk size that is best suitable to answer the questoin
        # Re-rank embeddings
        if llm_bundle.reranker is not None:
            # reranker_score, relevant_docs = llm_bundle.reranker.rerank(
            # original_query, all_relevant_docs
            # )
            reranker_score, relevant_docs = rerank(
                rerank_model=llm_bundle.reranker,
                query=original_query,
                docs=all_relevant_docs,
            )
            if reranker_score > current_reranker_score:
                logger.info("New most relevant query: {}", original_query)
                docs = relevant_docs
                current_reranker_score = reranker_score

        # logger.info(
        # f"Number of documents after stage 1 (sparse): {len(sparse_search_docs_ids)}"
        # )
        logger.info(
            f"Number of documents after stage 2 (dense + sparse): {len(all_relevant_docs)}"
        )
        logger.info(
            f"Re-ranker avg. scores for top 5 resuls, chunk size {chunk_size}: {reranker_score:.2f}"
        )

    len_ = 0

    for doc in docs:
        doc_length = len(doc.page_content)
        if len_ + doc_length < config.max_char_size:
            most_relevant_docs.append(doc)
            len_ += doc_length

    return most_relevant_docs, current_reranker_score



import hashlib
import urllib
import uuid
from pathlib import Path
from typing import List, Optional, Tuple

from loguru import logger
import pandas as pd

from llmsearch.config import Config, Document
from llmsearch.parsers.doc import docx_splitter
from llmsearch.parsers.markdown import markdown_splitter
from llmsearch.parsers.pdf import PDFSplitter
from llmsearch.parsers.unstructured import UnstructuredSplitter


HASH_BLOCKSIZE = 65536


class DocumentSplitter:
    def __init__(self, config: Config) -> None:

        # Custom splitters are configured here
        self._splitter_conf = {
            "md": markdown_splitter,
            "docx": docx_splitter,
            "doc": docx_splitter,
            "pdf": PDFSplitter(chunk_overlap=200).split_document,
        }

        # Fallback splitter unless custom splitter is specified
        self._fallback_splitter = UnstructuredSplitter().split_document

        self.document_path_settings = config.embeddings.document_settings
        self.chunk_sizes = config.embeddings.chunk_sizes

    def get_hashes(self) -> pd.DataFrame:
        hash_filename_mappings = []
        logger.info("Scanning hashes of the existing files.")

        for setting in self.document_path_settings:
            docs_path = Path(setting.doc_path)
            exclusion_paths = [str(e) for e in setting.exclude_paths]

            for scan_extension in setting.scan_extensions:
                extension = scan_extension

                # Create a list of document paths to process. Filter out paths in the exclusion list
                paths = [
                    p
                    for p in list(docs_path.glob(f"**/*.{extension}"))
                    if (not self.is_exclusion(p, exclusion_paths)) and (p.is_file())
                ]
                hashes = [
                    {"filename": str(path), "filehash": get_md5_hash(path)}
                    for path in paths
                ]
                hash_filename_mappings.extend(hashes)
        return pd.DataFrame(hash_filename_mappings)

    def split(
        self, restrict_filenames: Optional[List[str]] = None
    ) -> Tuple[List[Document], pd.DataFrame, pd.DataFrame]:
        """Splits documents based on document path settings

        Returns:
            List[Document]: List of documents
        """
        all_docs = []

        # Maps between file name and it's hash
        hash_filename_mappings = []

        # Mapping between hash and document ids
        hash_docid_mappings = []

        for setting in self.document_path_settings:
            passage_prefix = setting.passage_prefix
            docs_path = Path(setting.doc_path)
            documents_label = setting.label
            exclusion_paths = [str(e) for e in setting.exclude_paths]

            for scan_extension in setting.scan_extensions:
                extension = scan_extension
                for chunk_size in self.chunk_sizes:  # type: ignore
                    logger.info(f"Scanning path for extension: {extension}")

                    # Create a list of document paths to process. Filter out paths in the exclusion list
                    paths = [
                        p
                        for p in list(docs_path.glob(f"**/*.{extension}"))
                        if not self.is_exclusion(p, exclusion_paths)
                    ]

                    # Used when updating the index, we don't need to parse all files again
                    if restrict_filenames is not None:
                        logger.warning(
                            f"Restrict filenames specificed. Scanning at most {len(restrict_filenames)} files."
                        )
                        paths = [p for p in paths if str(p) in set(restrict_filenames)]

                    # Get splitter unless fallback is specified
                    splitter = self._splitter_conf.get(
                        extension, self._fallback_splitter
                    )

                    # Get additional parser setting for a given extension, if present
                    additional_parser_settings = setting.additional_parser_settings.get(
                        extension, dict()
                    )

                    (
                        docs,
                        hf_mappings,
                        hd_mappings,
                    ) = self._get_documents_from_custom_splitter(
                        document_paths=paths,
                        splitter_func=splitter,
                        max_size=chunk_size,
                        passage_prefix=passage_prefix,
                        label=documents_label,
                        **additional_parser_settings,
                    )

                    logger.info(f"Got {len(docs)} chunks for type: {extension}")
                    all_docs.extend(docs)
                    hash_filename_mappings.extend(hf_mappings)
                    hash_docid_mappings.extend(hd_mappings)

        all_hash_filename_mappings = pd.DataFrame(hash_filename_mappings)
        all_hash_docid_mappings = pd.concat(hash_docid_mappings, axis=0)

        return all_docs, all_hash_filename_mappings, all_hash_docid_mappings

    def is_exclusion(self, path: Path, exclusions: List[str]) -> bool:
        """Checks if path has parent folders in list of exclusions

        Args:
            path (Path): _description_
            exclusions (List[str]): List of exclusion folders

        Returns:
            bool: True if path is in list of exclusions
        """

        exclusion_paths = [Path(p) for p in exclusions]
        for ex_path in exclusion_paths:
            if ex_path in path.parents:
                logger.info(
                    f"Excluding path {path} from documents, as path parent path is excluded."
                )
                return True
        return False

    def _get_documents_from_custom_splitter(
        self,
        document_paths: List[Path],
        splitter_func,
        max_size,
        passage_prefix: str,
        label: str,
        **additional_kwargs,
    ) -> Tuple[List[Document], List[dict], List[pd.DataFrame]]:
        """Gets list of nodes from a collection of documents

        Examples: https://gpt-index.readthedocs.io/en/stable/guides/primer/usage_pattern.html
        """

        all_docs = []

        # Maps between file name and it's hash
        hash_filename_mappings = []

        # Mapping between hash and document ids
        hash_docid_mappings = []

        if passage_prefix:
            logger.info(f"Will add the following passage prefix: {passage_prefix}")

        for path in document_paths:
            logger.info(
                f"Processing path using custom splitter: {path}, chunk size: {max_size}"
            )

            # docs_data = splitter_func(text, max_size)
            filename = str(path)
            additional_kwargs.update({"filename": filename})
            docs_data = splitter_func(path, max_size, **additional_kwargs)
            file_hash = get_md5_hash(path)

            path = urllib.parse.quote(str(path))  # type: ignore
            logger.info(path)

            docs = [
                Document(
                    page_content=passage_prefix + d["text"],
                    metadata={
                        **d["metadata"],
                        **{
                            "source": str(path),
                            "chunk_size": max_size,
                            "document_id": str(uuid.uuid1()),
                            "label": label,
                        },
                    },
                )
                for d in docs_data
            ]

            for d in docs: # 2024/03/14 - unstructured fills page as None for some formats like csv
                if 'page' in d.metadata and d.metadata['page'] is None:
                    d.metadata['page'] = -1
            
            all_docs.extend(docs)

            # Add hash to filename mapping and hash to doc ids mapping
            hash_filename_mappings.append(dict(filename=filename, filehash=file_hash))

            df_hash_docid = (
                pd.DataFrame()
                .assign(docid=[d.metadata["document_id"] for d in docs])
                .assign(filehash=file_hash)
            )

            hash_docid_mappings.append(df_hash_docid)

        logger.info(f"Got {len(all_docs)} nodes.")
        return all_docs, hash_filename_mappings, hash_docid_mappings


def get_md5_hash(file_path: Path) -> str:
    hasher = hashlib.md5()

    with open(file_path, "rb") as file:
        buf = file.read(HASH_BLOCKSIZE)
        while buf:
            hasher.update(buf)
            buf = file.read(HASH_BLOCKSIZE)

    return hasher.hexdigest()



import gc
import shutil
import tqdm
from pathlib import Path
from typing import List, Optional, Tuple

from langchain_community.vectorstores import Chroma
from loguru import logger

from llmsearch.config import Config
from llmsearch.embeddings import VectorStore, get_embedding_model
from llmsearch.parsers.splitter import Document


class VectorStoreChroma(VectorStore):
    def __init__(self, persist_folder: str, config: Config):
        self._persist_folder = persist_folder
        self._config = config
        self._embeddings = get_embedding_model(config.embeddings.embedding_model)
        self.batch_size = 200  # Limitation of Chromadb (2023/09 v0.4.8) - can add only 41666 documents at once

        self._retriever = None
        self._vectordb = None

    @property
    def retriever(self):
        if self._retriever is None:
            self._retriever = self._load_retriever()
        return self._retriever

    @property
    def vectordb(self):
        if self._vectordb is None:
            self._vectordb = Chroma(
                persist_directory=self._persist_folder,
                embedding_function=self._embeddings,
            )
        return self._vectordb

    def unload(self):
        self._vectordb = None
        self._retriever = None

        gc.collect()

    def create_index_from_documents(
        self,
        all_docs: List[Document],
        clear_persist_folder: bool = True,
    ):
        if clear_persist_folder:
            pf = Path(self._persist_folder)
            if pf.exists() and pf.is_dir():
                logger.warning(f"Deleting the content of: {pf}")
                shutil.rmtree(pf)

        logger.info("Generating and persisting the embeddings..")

        vectordb = None
        for group in tqdm.tqdm(
            chunker(all_docs, size=self.batch_size),
            total=int(len(all_docs) / self.batch_size),
        ):
            ids = [d.metadata["document_id"] for d in group]
            if vectordb is None:
                vectordb = Chroma.from_documents(
                    documents=group,  # type: ignore
                    embedding=self._embeddings,
                    ids=ids,
                    persist_directory=self._persist_folder,  # type: ignore
                )
            else:
                vectordb.add_texts(
                    texts=[doc.page_content for doc in group],
                    embedding=self._embeddings,
                    ids=ids,
                    metadatas=[doc.metadata for doc in group],
                )
        logger.info("Generated embeddings. Persisting...")
        if vectordb is not None:
            vectordb.persist()
        vectordb = None

    def _load_retriever(self, **kwargs):
        # vectordb = Chroma(persist_directory=self._persist_folder, embedding_function=self._embeddings)
        return self.vectordb.as_retriever(**kwargs)

    def add_documents(self, docs: List[Document]):
        """Adds new documents to existing vectordb

        Args:
            docs (List[Document]): List of documents
        """

        logger.info(f"Adding embeddings for {len(docs)} documents")
        # vectordb = Chroma(persist_directory=self._persist_folder, embedding_function=self._embeddings)
        for group in tqdm.tqdm(
            chunker(docs, size=self.batch_size), total=int(len(docs) / self.batch_size)
        ):
            ids = [d.metadata["document_id"] for d in group]
            self.vectordb.add_texts(
                texts=[doc.page_content for doc in group],
                embedding=self._embeddings,
                ids=ids,
                metadatas=[doc.metadata for doc in group],
            )
        logger.info("Generated embeddings. Persisting...")
        self.vectordb.persist()

    def delete_by_id(self, ids: List[str]):
        logger.warning(f"Deleting {len(ids)} chunks.")
        # vectordb = Chroma(persist_directory=self._persist_folder, embedding_function=self._embeddings)
        self.vectordb.delete(ids=ids)
        self.vectordb.persist()

    def get_documents_by_id(self, document_ids: List[str]) -> List[Document]:
        """Retrieves documents by ids

        Args:
            document_ids (List[str]): list of document ids

        Returns:
            List[Document]: list of documents belonging to document_ids
        """

        results = self.retriever.vectorstore.get(ids=document_ids, include=["metadatas", "documents"])  # type: ignore
        docs = [
            Document(page_content=d, metadata=m)
            for d, m in zip(results["documents"], results["metadatas"])
        ]
        return docs

    def similarity_search_with_relevance_scores(
        self, query: str, k: int, filter: Optional[dict]
    ) -> List[Tuple[Document, float]]:
        # If there are multiple key-value pairs, combine using AND rule - the syntax is chromadb specific
        if isinstance(filter, dict) and len(filter) > 1:
            filter = {"$and": [{key: {"$eq": value}} for key, value in filter.items()]}
            print("Filter = ", filter)

        return self.retriever.vectorstore.similarity_search_with_relevance_scores(
            query, k=self._config.semantic_search.max_k, filter=filter
        )  # type: ignore


def chunker(seq, size):
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))



import os
from abc import ABC, abstractmethod
from typing import List, Tuple

import pandas as pd
from langchain_community.embeddings import (
    HuggingFaceEmbeddings,
    HuggingFaceInstructEmbeddings,
    SentenceTransformerEmbeddings,
)
from loguru import logger

from llmsearch.config import Config, Document, EmbeddingModel, EmbeddingModelType
from llmsearch.parsers.splitter import DocumentSplitter
from llmsearch.splade import SparseEmbeddingsSplade

MODELS = {
    EmbeddingModelType.instruct: HuggingFaceInstructEmbeddings,
    EmbeddingModelType.sentence_transformer: SentenceTransformerEmbeddings,
    EmbeddingModelType.huggingface: HuggingFaceEmbeddings,
}


class VectorStore(ABC):
    @abstractmethod
    def create_index_from_documents(
        self, all_docs: List[Document], clear_persist_folder: bool = True
    ):
        pass

    @abstractmethod
    def get_documents_by_id(self, docuemt_ids: List[str]):
        pass

    @property
    @abstractmethod
    def retriever(self):
        pass

    @abstractmethod
    def similarity_search_with_relevance_scores(self, query: str, k: int, filter: dict):
        pass

    @abstractmethod
    def delete_by_id(self, ids: List[str]):
        pass

    @abstractmethod
    def add_documents(self, docs: List[Document]):
        pass


class EmbeddingsHashNotExistError(Exception):
    """Raised when embeddings can't be founrd"""


def get_embedding_model(config: EmbeddingModel):
    """Loads an embedidng model

    Args:
        config (EmbeddingModel): Configuration for the embedding model

    Raises:
        TypeError: if model is unsupported
    """

    logger.info(f"Embedding model config: {config}")
    model_type = MODELS.get(config.type, None)

    if model_type is None:
        raise TypeError(f"Unknown model type. Got {config.type}")

    return model_type(model_name=config.model_name, **config.additional_kwargs)


def create_embeddings(config: Config, vs: VectorStore):
    splitter = DocumentSplitter(config)
    all_docs, all_hash_filename_mappings, all_hash_docid_mappings = splitter.split()

    vs.create_index_from_documents(all_docs=all_docs)

    splade = SparseEmbeddingsSplade(config)
    splade.generate_embeddings_from_docs(docs=all_docs)

    save_document_hashes(config, all_hash_filename_mappings, all_hash_docid_mappings)
    logger.info("ALL DONE.")


def update_embeddings(config: Config, vs: VectorStore) -> dict:
    splitter = DocumentSplitter(config)
    new_hashes_df = splitter.get_hashes()
    try:
        file_hashes_fn, docid_hash_fn = get_hash_mapping_filenames(config)
        existing_fn_hash_mappings = pd.read_parquet(path=file_hashes_fn)
    except FileNotFoundError:
        raise EmbeddingsHashNotExistError(
            "Hash file don't exist, please re-create the index."
        )
    existing_docid_hash_mappings = pd.read_parquet(path=docid_hash_fn)
    stats = {
        "original_n_files": len(existing_fn_hash_mappings),
        "updated_n_files": 0,
        "scanned_files": 0,
        "scanned_chunks": 0,
        "changed_files": 0,
        "changed_chunks": 0,
        "deleted_files": 0,
        "deleted_chunks": 0,
    }

    changed_or_new_df, changed_df, deleted_df = get_changed_or_new_files(
        new_hashes_df, existing_fn_hash_mappings
    )
    logger.debug("===== Rescanned files ======")
    logger.debug(changed_or_new_df)
    logger.debug("===== Changed files ======")
    logger.debug(changed_df)
    logger.debug("===== Deleted files ======")
    logger.debug(deleted_df)

    if len(changed_or_new_df) == 0 and len(changed_df) == 0 and len(deleted_df) == 0:
        logger.info("The index is up to date. Exiting.")
        return stats

    splade = SparseEmbeddingsSplade(config=config)
    splade.load()

    # Delete chunks belonging to changed documents that exist in both new and old index
    if len(changed_df) > 0:
        changed_doc_ids = existing_docid_hash_mappings.loc[
            existing_docid_hash_mappings["filehash"].isin(changed_df["filehash"]),
            "docid",
        ].tolist()

        logger.info(
            "Removing chunks from vectorstore belonging to changed documents..."
        )
        vs.delete_by_id(ids=changed_doc_ids)
        logger.info("Removing mappings belonging to changed documents...")
        existing_fn_hash_mappings, existing_docid_hash_mappings = delete_mappings(
            existing_fn_hash_mappings, existing_docid_hash_mappings, changed_df
        )

        # Delete splade embeddings
        splade.delete_by_ids(delete_ids=changed_doc_ids)
        stats["changed_files"] = len(changed_df)
        stats["changed_chunks"] = len(changed_doc_ids)

    # Delete chunks belonging to deleted documents that exist in only in old index
    if len(deleted_df) > 0:
        logger.info(
            "Removing chunks from vectorstore belonging to deleted documents..."
        )
        deleted_doc_ids = existing_docid_hash_mappings.loc[
            existing_docid_hash_mappings["filehash"].isin(deleted_df["filehash"]),
            "docid",
        ].tolist()
        vs.delete_by_id(ids=deleted_doc_ids)

        logger.info("Removing mappings belonging to deleted documents...")
        existing_fn_hash_mappings, existing_docid_hash_mappings = delete_mappings(
            existing_fn_hash_mappings, existing_docid_hash_mappings, deleted_df
        )

        # Delete splade embeddings
        splade.delete_by_ids(delete_ids=deleted_doc_ids)
        stats["deleted_files"] = len(deleted_df)
        stats["deleted_chunks"] = len(deleted_doc_ids)

    # Rescan new and changed documents and add them to vectorstore
    if len(changed_or_new_df) > 0:
        splitter = DocumentSplitter(config)

        new_docs, new_fn_hash_mappings, new_docid_hash_mappings = splitter.split(
            restrict_filenames=changed_or_new_df.loc[:, "filename"].tolist()
        )

        existing_fn_hash_mappings, existing_docid_hash_mappings = add_mappings(
            existing_fn_hash_mappings,
            new_fn_hash_mappings,
            existing_docid_hash_mappings,
            new_docid_hash_mappings,
        )
        vs.add_documents(new_docs)

        splade.add_embeddings(new_docs)
        stats["scanned_files"] = len(changed_or_new_df)
        stats["scanned_chunks"] = len(new_docs)

    stats["updated_n_files"] = len(existing_fn_hash_mappings)
    # Save changed mappings
    save_document_hashes(
        config, existing_fn_hash_mappings, existing_docid_hash_mappings
    )
    return stats


def delete_mappings(
    existing_fn_hash_mappings: pd.DataFrame,
    existing_docid_hash_mappings: pd.DataFrame,
    changed_df,
):
    # In the existing mappings, delete all rows belonging to changed filenames
    mask_delete_fn_hash = existing_fn_hash_mappings.loc[:, "filename"].isin(
        changed_df.loc[:, "filename"]
    )
    updated_fn_hash_mappings = existing_fn_hash_mappings.loc[~mask_delete_fn_hash, :]

    # Update existings mappings with new fn mappings
    mask_delete_docid_hash = existing_docid_hash_mappings.loc[:, "filehash"].isin(
        changed_df.loc[:, "filehash"]
    )
    updated_docid_hash_mappings = existing_docid_hash_mappings.loc[
        ~mask_delete_docid_hash, :
    ]
    return updated_fn_hash_mappings, updated_docid_hash_mappings


def add_mappings(
    existing_fn_hash_mappings: pd.DataFrame,
    new_fn_hash_mappings: pd.DataFrame,
    existing_docid_hash_mappings: pd.DataFrame,
    new_docid_hash_mappings: pd.DataFrame,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Adds new hash and docid mappings to the existing ones.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: updated filename to hash mapping df,
            updated docid to hash mapping df
    """

    # Update existings mappings with new fn mappings
    updated_fn_hash_mappings = pd.concat(
        [existing_fn_hash_mappings, new_fn_hash_mappings], axis=0
    )

    updated_docid_hash_mappings = pd.concat(
        [existing_docid_hash_mappings, new_docid_hash_mappings], axis=0
    )
    return updated_fn_hash_mappings, updated_docid_hash_mappings


def get_changed_or_new_files(
    new_hashes_df: pd.DataFrame, existing_hashes_df: pd.DataFrame
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Calculates changes between new file hashes and existing file hashes

    Args:
        new_hashes_df (pd.DataFrame): Dataframe containing newly scanned file hashes
        existing_hashes_df (pd.DataFrame): Dataframe containing existing file hashes


    Returns:
        Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: Dataframes containing:
            changed or new files - these files will need to be (re)scanned.
            changed_files - these files to be rescanned, and old docids to be deleted from the vector store
            deleted - docids belonging to these files have to be deleted from the vector store.
    """

    merged_df = new_hashes_df.merge(
        existing_hashes_df, on=["filehash", "filename"], how="outer", indicator=True
    )

    # Rows with left_only indicators contain files that are either changed or new in the new scan
    # These files are to be re-scanned
    changed_or_new = merged_df.loc[lambda df: df["_merge"] == "left_only"]

    duplicated_filenames_mask = merged_df.loc[:, "filename"].duplicated()
    changed = merged_df.loc[duplicated_filenames_mask, :]

    # Deleted files - mask is "right_only" (existed only in the previous df), but also not in the changed files
    deleted_mask = (merged_df["_merge"] == "right_only") & (
        ~merged_df["filename"].isin(changed["filename"])
    )
    deleted = merged_df.loc[deleted_mask, :]

    return changed_or_new, changed, deleted


def get_hash_mapping_filenames(
    config: Config,
    file_to_hash_fn: str = "file_hash_mappings.snappy.parquet",
    docid_to_hash_fn="docid_hash_mappings.snappy.parquet",
) -> Tuple[str, str]:
    """Returns filenames to store hashes"""

    file_hashes_fn = os.path.join(config.embeddings.embeddings_path, file_to_hash_fn)
    docid_hashes_fn = os.path.join(config.embeddings.embeddings_path, docid_to_hash_fn)
    return file_hashes_fn, docid_hashes_fn


def save_document_hashes(
    config: Config,
    all_hash_filename_mappings: pd.DataFrame,
    all_hash_docid_mappings: pd.DataFrame,
) -> None:
    """Saves hashes of the processed files and mappings from hashes to individual chunks

    Args:
        config (Config): Instance of Config
        all_hash_filename_mappings (pd.DataFrame): Dataframe containing mappings between filename and file hash
        all_hash_docid_mappings (pd.DataFrame): Dataframe contains mappings between file hash and document ids
    """

    file_hashes_fn, docid_hashes_fn = get_hash_mapping_filenames(config)

    logger.info(f"Saving file hashes mappings to {file_hashes_fn}")
    all_hash_filename_mappings.to_parquet(
        file_hashes_fn, compression="snappy", index=False
    )
    logger.info(
        f"Distinct hashes generated: {len(all_hash_filename_mappings['filehash'].unique())}"
    )

    logger.info(f"Saving document id to hash mappings to {docid_hashes_fn}")
    all_hash_docid_mappings.to_parquet(
        docid_hashes_fn, compression="snappy", index=False
    )
    logger.info(f"Hash to document id mappings: {len(all_hash_docid_mappings)}")



import langchain
from dotenv import load_dotenv
from termcolor import cprint

from llmsearch.config import Config, ResponseModel
from llmsearch.process import get_and_parse_response
from llmsearch.utils import LLMBundle

load_dotenv()
langchain.debug = True


def print_llm_response(output: ResponseModel):
    print("\n============= SOURCES ==================")
    for source in output.semantic_search[::-1]:
        source.metadata.pop("source")
        cprint(source.chunk_link, "blue")
        cprint(source.metadata, "cyan")
        print("******************* BEING EXTRACT *****************")
        print(f"{source.chunk_text}\n")
    print("\n============= RESPONSE =================")
    cprint(output.response, "red")
    print("------------------------------------------")


def qa_with_llm(llm_bundle: LLMBundle, config: Config):
    while True:
        question = input("\nENTER QUESTION >> ")
        output = get_and_parse_response(
            query=question,
            llm_bundle=llm_bundle,
            config=config,
        )
        print_llm_response(output)



from llmsearch.models.abstract import AbstractLLMModel
from llmsearch.models.config import OpenAIModelConfig
from langchain_community.chat_models import ChatOpenAI

# from langchain_openai import ChatOpenAI


class OpenAIModel(AbstractLLMModel):
    def __init__(self, config: OpenAIModelConfig) -> None:
        super().__init__(prompt_template=config.prompt_template)
        self.config = config

    @property
    def model(self):
        return ChatOpenAI(**self.config.model_kwargs)



from enum import Enum, auto
from pathlib import Path
from typing import List, Union

from loguru import logger
from unstructured.documents.elements import NarrativeText, Text, Title

from unstructured.partition.auto import partition


class UnstructuredSplitter:
    def __init__(self) -> None:
        self.partition_function = partition
        self.supported_elements = (NarrativeText, Text, Title)

    def split_document(
        self, document_path: Union[str, Path], max_size: int, **kwargs
    ) -> List[dict]:
        logger.info(f"Partitioning document: {document_path}")

        elements = self.partition_function(filename=str(document_path))

        logger.info("Combining document chunks...")

        all_chunks = []

        current_chunk = ""
        current_page = 1e8
        for el in elements:
            if not isinstance(el, self.supported_elements):
                continue
            text = "\n" + el.text

            # If element's text is larger than chunk size, split by characters
            if len(text) >= max_size - 1:
                # Flush the current chunk
                all_chunks.append(
                    {"text": current_chunk, "metadata": {"page": current_page}}
                )

                texts = text_split(text, max_size)
                logger.info(f"Sub-splitting large chunk. Got {len(texts)} sub-chunks.")
                for t in texts:
                    all_chunks.append(
                        {
                            "text": "\n" + t,
                            "metadata": {"page": el.metadata.page_number},
                        }
                    )

                current_chunk = ""
                current_page = 1e8

            # if element's text doesn't fit into current chunk - flush the current chunk and create a new chunk
            elif len(current_chunk + text) >= max_size:
                all_chunks.append(
                    {"text": current_chunk, "metadata": {"page": current_page}}
                )
                logger.info(
                    f"Flushing chunk. Length: {len(current_chunk)}, page: {el.metadata.page_number}"
                )
                current_chunk = text
                current_page = el.metadata.page_number

            # Otherwise, add element's text to current chunk, without re-assigning the page number
            else:
                current_chunk += text
                # If page number is None, assign it to 1
                current_page = current_page or 1e8
                metadata_page_number = el.metadata.page_number or current_page
                current_page = min(current_page, metadata_page_number)

        return all_chunks


def text_split(s, w):
    return [s[i : i + w] for i in range(0, len(s), w)]



# from llmsearch.models.abstract import AbstractLLMModel
# import auto_gptq
# from llmsearch.models.config import AutoGPTQModelConfig
# import torch

# import transformers
# from langchain.llms import HuggingFacePipeline


# class AutoGPTQModel(AbstractLLMModel):
#     def __init__(self, config: AutoGPTQModelConfig) -> None:
#         super().__init__(prompt_template=config.prompt_template)
#         self.config = config

#     @property
#     def model(self):
#         if self.config.device == "auto":
#             device = (
#                 f"cuda:{torch.cuda.current_device()}"
#                 if torch.cuda.is_available()
#                 else "cpu"
#             )
#         else:
#             device = self.config.device

#         tokenizer = transformers.AutoTokenizer.from_pretrained(
#             self.config.model_folder, use_fast=True, device=device
#         )

#         model = auto_gptq.AutoGPTQForCausalLM.from_quantized(
#             self.config.model_folder,
#             use_safetensors=self.config.use_safetensors,
#             trust_remote_code=self.config.trust_remote_code,
#             device=device,
#             **self.config.model_kwargs,
#         )

#         p = transformers.pipeline(
#             "text-generation",
#             model=model,
#             tokenizer=tokenizer,
#             **self.config.pipeline_kwargs,
#         )

#         hf_pipeline = HuggingFacePipeline(pipeline=p)
#         return hf_pipeline



from typing import Dict, Generator, Optional
from langchain.callbacks.manager import CallbackManager
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler

from llmsearch.models.abstract import AbstractLLMModel
from llmsearch.models.config import LlamaModelConfig
from typing import Any, List

from langchain.callbacks.manager import CallbackManagerForLLMRun
from langchain.llms.base import LLM
from llama_cpp import Llama
from loguru import logger

# class LlamaModel(AbstractLLMModel):
#     def __init__(self, config: LlamaModelConfig):
#         super().__init__(prompt_template=config.prompt_template)
#         self.config = config

#     @property
#     def model(self):

#         callback_manager = CallbackManager([StreamingStdOutCallbackHandler()])
#         llm = LlamaCpp(
#             model_path=str(self.config.model_path),
#             callback_manager=callback_manager, verbose=True, **self.config.model_kwargs
#         )
#         return llm


# values["client"] = Llama(model_path, **model_params)


class CustomLlamaLangChainModel(LLM):
    @classmethod
    def from_parameters(cls, model_path, model_init_kwargs, model_kwargs, **kwargs):
        cls.model = Llama(model_path=str(model_path), **model_init_kwargs)
        cls.model_kwargs = model_kwargs
        cls.model_path = model_path
        cls.streaming = True
        return cls(**kwargs)

    def __del__(self):
        self.model.__del__()

    @property
    def _llm_type(self) -> str:
        return "custom"

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
    ) -> str:
        if stop is not None:
            raise ValueError("stop kwargs are not permitted.")

        if self.streaming:
            # If streaming is enabled, we use the stream
            # method that yields as they are generated
            # and return the combined strings from the first choices's text:
            combined_text_output = ""
            for token in self.stream(prompt=prompt, stop=stop, run_manager=run_manager):
                combined_text_output += token["choices"][0]["text"]
            return combined_text_output
        else:
            result = self.model(prompt=prompt, **self.model_kwargs)
            return result["choices"][0]["text"]

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Get the identifying parameters."""
        return {**{"model_path": self.model_path}, **self.model_kwargs}

    def stream(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
    ) -> Generator[Dict, None, None]:
        """Yields results objects as they are generated in real time.

        BETA: this is a beta feature while we figure out the right abstraction.
        Once that happens, this interface could change.

        It also calls the callback manager's on_llm_new_token event with
        similar parameters to the OpenAI LLM class method of the same name.

        Args:
            prompt: The prompts to pass into the model.
            stop: Optional list of stop words to use when generating.

        Returns:
            A generator representing the stream of tokens being generated.

        Yields:
            A dictionary like objects containing a string token and metadata.
            See llama-cpp-python docs and below for more.

        Example:
            .. code-block:: python

                from langchain.llms import LlamaCpp
                llm = LlamaCpp(
                    model_path="/path/to/local/model.bin",
                    temperature = 0.5
                )
                for chunk in llm.stream("Ask 'Hi, how are you?' like a pirate:'",
                        stop=["'","\n"]):
                    result = chunk["choices"][0]
                    print(result["text"], end='', flush=True)

        """
        result = self.model(prompt=prompt, stream=True, **self.model_kwargs)
        for chunk in result:
            token = chunk["choices"][0]["text"]
            log_probs = chunk["choices"][0].get("logprobs", None)
            if run_manager:
                run_manager.on_llm_new_token(
                    token=token, verbose=self.verbose, log_probs=log_probs
                )
            yield chunk


class LlamaModel(AbstractLLMModel):
    def __init__(self, config: LlamaModelConfig):
        super().__init__(prompt_template=config.prompt_template)
        self.config = config
        self._model = None

    @property
    def model(self):
        if not self._model:
            logger.info("Loading model...")
            callback_manager = CallbackManager([StreamingStdOutCallbackHandler()])

            logger.info("Initializing LLAmaCPP model...")
            logger.info(self.config.model_init_params)
            model_path = self.config.model_path
            model_kwargs = self.config.model_kwargs
            model_init_kwargs = self.config.model_init_params

            self._model = CustomLlamaLangChainModel.from_parameters(
                model_path=str(model_path),
                model_kwargs=model_kwargs,
                model_init_kwargs=model_init_kwargs,
                callback_manager=callback_manager,
                verbose=True,
            )
        return self._model



import json
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import docx
from docx.document import Document as doctwo
from docx.oxml.table import CT_Tbl
from docx.oxml.text.paragraph import CT_P
from docx.table import Table, _Cell
from docx.text.paragraph import Paragraph
from loguru import logger


@dataclass
class Heading:
    level: int
    text: str


@dataclass
class HeadingSequence:
    headings: List[Heading] = field(default_factory=list)

    def add(self, text: str, level: int):
        """Adds new headings to the heading sequence

        Args:
            text (str): text of the heading
            level (int): level of the heading (0 - most significant, e.g. title)
        """

        current_heading = Heading(level, text.strip())

        while self.headings and self.headings[-1].level >= current_heading.level:
            self.headings.pop()
        self.headings.append(current_heading)

    @property
    def path(self) -> str:
        s = [t.text for t in self.headings]
        return "/".join(s)


## Based on langchain's splitter


class RecursiveCharacterTextSplitter:
    """Splitting text by recursively look at characters.

    Recursively tries to split by different characters to find one
    that works.
    """

    def __init__(
        self,
        chunk_size: int,
        length_function: Callable[[str], int] = len,
        chunk_overlap: int = 200,
        separators: Optional[List[str]] = None,
        keep_separator: bool = True,
        is_separator_regex: bool = False,
        **kwargs: Any,
    ) -> None:
        """Create a new TextSplitter."""
        #        super().__init__(keep_separator=keep_separator, **kwargs)
        self._keep_separator = keep_separator
        self._separators = separators or ["\n\n", "\n", " ", ""]
        self._is_separator_regex = is_separator_regex
        self._chunk_size = chunk_size
        self._length_function = length_function
        self._chunk_overlap = chunk_overlap

    def _split_text(self, text: str, separators: List[str]) -> List[str]:
        """Split incoming text and return chunks."""
        final_chunks = []
        # Get appropriate separator to use
        separator = separators[-1]
        new_separators = []
        for i, _s in enumerate(separators):
            _separator = _s if self._is_separator_regex else re.escape(_s)
            if _s == "":
                separator = _s
                break
            if re.search(_separator, text):
                separator = _s
                new_separators = separators[i + 1 :]
                break

        _separator = separator if self._is_separator_regex else re.escape(separator)
        splits = _split_text_with_regex(text, _separator, self._keep_separator)

        # Now go merging things, recursively splitting longer texts.
        _good_splits = []
        _separator = "" if self._keep_separator else separator
        for s in splits:
            if self._length_function(s) < self._chunk_size:
                _good_splits.append(s)
            else:
                if _good_splits:
                    merged_text = self._merge_splits(_good_splits, _separator)
                    final_chunks.extend(merged_text)
                    _good_splits = []
                if not new_separators:
                    final_chunks.append(s)
                else:
                    other_info = self._split_text(s, new_separators)
                    final_chunks.extend(other_info)
        if _good_splits:
            merged_text = self._merge_splits(_good_splits, _separator)
            final_chunks.extend(merged_text)
        return final_chunks

    def split_text(self, text: str) -> List[str]:
        return self._split_text(text, self._separators)

    def _join_docs(self, docs: List[str], separator: str) -> Optional[str]:
        text = separator.join(docs)
        text = text.strip()
        if text == "":
            return None
        else:
            return text

    def _merge_splits(self, splits: Iterable[str], separator: str) -> List[str]:
        # We now want to combine these smaller pieces into medium size
        # chunks to send to the LLM.
        separator_len = self._length_function(separator)

        docs = []
        current_doc: List[str] = []
        total = 0
        for d in splits:
            _len = self._length_function(d)
            if (
                total + _len + (separator_len if len(current_doc) > 0 else 0)
                > self._chunk_size
            ):
                if total > self._chunk_size:
                    logger.warning(
                        f"Created a chunk of size {total}, "
                        f"which is longer than the specified {self._chunk_size}"
                    )
                if len(current_doc) > 0:
                    doc = self._join_docs(current_doc, separator)
                    if doc is not None:
                        docs.append(doc)
                    # Keep on popping if:
                    # - we have a larger chunk than in the chunk overlap
                    # - or if we still have any chunks and the length is long
                    while total > self._chunk_overlap or (
                        total + _len + (separator_len if len(current_doc) > 0 else 0)
                        > self._chunk_size
                        and total > 0
                    ):
                        total -= self._length_function(current_doc[0]) + (
                            separator_len if len(current_doc) > 1 else 0
                        )
                        current_doc = current_doc[1:]
            current_doc.append(d)
            total += _len + (separator_len if len(current_doc) > 1 else 0)
        doc = self._join_docs(current_doc, separator)
        if doc is not None:
            docs.append(doc)
        return docs


def _split_text_with_regex(
    text: str, separator: str, keep_separator: bool
) -> List[str]:
    # Now that we have the separator, split the text
    if separator:
        if keep_separator:
            # The parentheses in the pattern keep the delimiters in the result.
            _splits = re.split(f"({separator})", text)
            splits = [_splits[i] + _splits[i + 1] for i in range(1, len(_splits), 2)]
            if len(_splits) % 2 == 0:
                splits += _splits[-1:]
            splits = [_splits[0]] + splits
        else:
            splits = re.split(separator, text)
    else:
        splits = list(text)
    return [s for s in splits if s != ""]


##This function extracts the tables and paragraphs from the document object
def iter_block_items(parent):
    """
    Yield each paragraph and table child within *parent*, in document order.
    Each returned value is an instance of either Table or Paragraph. *parent*
    would most commonly be a reference to a main Document object, but
    also works for a _Cell object, which itself can contain paragraphs and tables.
    """

    if isinstance(parent, doctwo):
        parent_elm = parent.element.body
    elif isinstance(parent, _Cell):
        parent_elm = parent._tc
    else:
        raise ValueError("something's not right")

    for child in parent_elm.iterchildren():
        if isinstance(child, CT_P):
            yield Paragraph(child, parent)
        elif isinstance(child, CT_Tbl):
            yield Table(child, parent)


def parse_table(table: Table) -> List[dict]:
    """Parses a single table recursively. Supports nested tables.

    Args:
        table (Table): Table element to split

    Returns:
        list: List of dictionaries containing rows of the tables
    """

    headers = get_table_header(table)

    out_list = list()
    current_dict = dict()

    for row in table.rows[1:]:
        for header, cell in zip(headers, row.cells):
            if cell.tables:
                current_dict[header] = []
                current_dict[header].append(cell.text)
            else:
                current_dict[header] = cell.text

            for i, nested_table in enumerate(cell.tables):
                r = parse_table(nested_table)
                if r:
                    current_dict[header].append(r)

        out_list.append(current_dict)
        current_dict = dict()

    return out_list


def get_table_header(table: Table):
    header = tuple([cell.text for cell in table.rows[0].cells])
    return header


def docx_splitter(
    path: Union[str, Path], max_chunk_size: int, **additional_splitter_setting
):
    doc = docx.Document(path)

    hs = HeadingSequence()

    out_chunks = []
    current_chunk = ""

    # Splitter for text paragraphs
    paragraph_splitter = RecursiveCharacterTextSplitter(
        chunk_size=max_chunk_size, separators=["\n\n", "\n", ". ", " "]
    )

    # Splitter for json
    json_splitter = RecursiveCharacterTextSplitter(
        chunk_size=max_chunk_size,
        separators=[r"\{.*?\}", r'".*?"\s*:\s*".*?"', r"\n\n", r"\n", r"\. ", r"\."],
        is_separator_regex=True,
        chunk_overlap=0,
    )

    # Iterate over all elements in the document

    for el in iter_block_items(doc):
        if isinstance(el, Paragraph):
            p_text = el.text

            # If text of the paragraph is empty, continue
            if not p_text.strip():
                continue

            # Store headings for metadata
            # if "title" in el.style.name.lower():
            #    hs.add(p_text, level=0)
            elif "heading" in el.style.name.lower():
                heading_level = int(el.style.name.lower().split(" ")[-1])
                hs.add(p_text, heading_level)

            current_chunk = add_or_split(
                p_text,
                paragraph_splitter,
                current_chunk,
                max_chunk_size,
                out_chunks,
                hs,
                type="paragraph",
            )

        elif isinstance(el, Table):
            t_json = json.dumps(parse_table(el))
            current_chunk = add_or_split(
                t_json,
                json_splitter,
                current_chunk,
                max_chunk_size,
                out_chunks,
                hs,
                type="tabular data in json format.",
            )

    return out_chunks


def add_or_split(
    text,
    splitter,
    current_chunk: str,
    chunk_size: int,
    out_chunks: List[Dict[str, str]],
    hs: HeadingSequence,
    type: str,
) -> str:
    """Adds or splits text to out chunks, together with additional metadata

    Args:
        text (str): Text candidate to add
        splitter (Instance of the splitter): Instance of the splitter
        current_chunk (str): Current buffer (not flushed yet)
        chunk_size (int): Target chunk size
        out_chunks (List[str]): List of all historical chunks
        hs (HeadingSequence): Holding the metadata for the text

    Returns:
        str: current chunk (not flushed)
    """
    additional_metadata = {"type": type}

    # Case 1- length of the next paragraph > chunk_size, flulsh the current and split the next
    if len(text) >= chunk_size:
        out_chunks.append(add_metadata(hs, current_chunk, additional_metadata))
        current_chunk = ""

        for ch in splitter.split_text(text):
            out_chunks.append(add_metadata(hs, ch, additional_metadata))

    # Case 2 - next paragraph is smaller than chunk size, but can't be added to the current chunk
    elif len(current_chunk) + len(text) >= chunk_size:
        out_chunks.append(add_metadata(hs, current_chunk, additional_metadata))

        current_chunk = text

    # Case 3 - current chunk is small enough
    else:
        current_chunk += text

    return current_chunk


def add_metadata(
    hs: HeadingSequence, text: str, additional_metadata: Optional[dict] = None
) -> Dict[str, str]:
    metadata_s = ""

    additional_metadata = {} if additional_metadata is None else additional_metadata
    additional_metadata.update({"topic": f"{hs.path}"})

    for k, v in additional_metadata.items():
        if v:
            metadata_s += f"{k}: {v}\n"

    metadata = f"Metadata applicable to the next chunk of text delimited by five stars:\n<< METADATA\n{metadata_s}METADATA\n\n"
    t = metadata + "*****\n" + text + "\n*****"
    chunk = {"text": t, "metadata": {"heading": hs.path}}
    return chunk


if __name__ == "__main__":
    fn = "/home/lacpd1/projects/doc-parsing/doc1.docx"
    parsed = docx_splitter(fn, max_chunk_size=1024)

    for p in parsed:
        print("-------------")
        print(len(p["text"]))
        print(p["text"])



import os

from langchain_community.chat_models import AzureChatOpenAI

from llmsearch.models.abstract import AbstractLLMModel
from llmsearch.models.config import AzureOpenAIModelConfig


class AzureOpenAIModel(AbstractLLMModel):
    def __init__(self, config: AzureOpenAIModelConfig) -> None:
        super().__init__(prompt_template=config.prompt_template)
        self.config = config

    @property
    def model(self):
        os.environ["OPENAI_API_TYPE"] = self.config.openai_api_type
        os.environ["OPENAI_API_BASE"] = self.config.openai_api_base
        os.environ["OPENAI_API_VERSION"] = self.config.openai_api_version

        return AzureChatOpenAI(
            deployment_name=self.config.deployment_name,
            model=self.config.model_name,
            **self.config.model_kwargs
        )



from pathlib import Path
from typing import List, Union

import fitz
from langchain.text_splitter import CharacterTextSplitter
from loguru import logger


class PDFSplitter:
    def __init__(self, chunk_overlap: int = 200) -> None:
        """Splits pdf documents

        Args:
            chunk_overlap (int, optional): Overlap in characters. Defaults to 200.
        """
        self.chunk_overlap = chunk_overlap

    def split_document(
        self, document_path: Union[str, Path], max_size: int, **kwargs
    ) -> List[dict]:
        """Splits pdf documents into multiple chunks

        Args:
            document_path (Union[str, Path]): Path to a single PDF document
            max_size (int): Max chunk size in characters.

        Returns:
            List[dict]: List of dictionaries containing parsed documents
        """

        logger.info(f"Partitioning document: {document_path}")

        all_chunks = []
        splitter = CharacterTextSplitter(
            separator="\n",
            keep_separator=True,
            chunk_size=max_size,
            chunk_overlap=self.chunk_overlap,
        )

        doc = fitz.open(document_path)
        current_text = ""
        for page in doc:
            text = page.get_text("block")

            if len(text) > max_size:
                all_chunks.append(
                    {"text": current_text, "metadata": {"page": page.number}}
                )
                chunks = splitter.split_text(text)
                for chunk in chunks:
                    logger.info(
                        f"Flushing chunk. Length: {len(chunk)}, page: {page.number}"
                    )
                    all_chunks.append(
                        {"text": chunk, "metadata": {"page": page.number}}
                    )
                current_text = ""

            elif len(current_text + text) >= max_size:
                if current_text != "":
                    all_chunks.append(
                        {"text": current_text, "metadata": {"page": page.number}}
                    )
                logger.info(
                    f"Flushing chunk. Length: {len(current_text)}, page: {page.number}"
                )
                current_text = text

            # Otherwise, add element's text to current chunk, without re-assigning the page number
            else:
                current_text += text

        # Filter out empty docs
        all_chunks = [
            chunk for chunk in all_chunks if chunk["text"].strip().replace(" ", "")
        ]
        return all_chunks



import uuid
from datetime import datetime

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    DateTime,
    Float,
    ForeignKey,
    Integer,
    String,
    Text,
)
from sqlalchemy.orm import relationship

from llmsearch.database.config import Base


def create_uuid() -> str:
    return str(uuid.uuid4())


class ResponseInteraction(Base):
    __tablename__ = "interactions"

    response_id = Column(String, primary_key=True, index=True)
    response_timestamp = Column(DateTime, default=datetime.utcnow)
    question_text = Column(Text, nullable=False)
    response_text = Column(Text, nullable=False)
    config_id = Column(String, ForeignKey("configs.config_id"))
    average_score = Column(Float, nullable=False)
    is_positive = Column(Boolean, nullable=True, default=None)
    feedback_text = Column(Text, nullable=True, default="")

    sources = relationship("InteractionSourcesBridge")


class InteractionSourcesBridge(Base):
    __tablename__ = "interactsourcebridge"

    record_id = Column(Integer, primary_key=True, autoincrement=True)
    response_id = Column(String, ForeignKey("interactions.response_id"), index=True)
    source_id = Column(String, ForeignKey("sources.source_id"), index=True)


class Sources(Base):
    __tablename__ = "sources"

    source_id = Column(String, primary_key=True, index=True)
    text_blob = Column(Text, nullable=False)
    text_link = Column(Text, nullable=False)
    rank_score = Column(Float, nullable=False)
    additional_metadata = Column(JSON)

    interactions = relationship("InteractionSourcesBridge")


class Config(Base):
    __tablename__ = "configs"

    config_id = Column(String, primary_key=True, index=True, default=create_uuid)
    config_hash = Column(String, nullable=False)
    config_json = Column(JSON, nullable=False)



from abc import ABC, abstractmethod
from typing import Optional

from langchain.prompts import PromptTemplate


class AbstractLLMModel(ABC):
    def __init__(self, prompt_template: Optional[str] = None) -> None:
        self.prompt_template = prompt_template

    @property
    @abstractmethod
    def model(self):
        raise NotImplementedError

    @property
    def prompt(self) -> Optional[PromptTemplate]:
        if self.prompt_template:
            return PromptTemplate(
                input_variables=["context", "question"], template=self.prompt_template
            )
        return None



import argparse
import gc
import os
from io import StringIO
from pathlib import Path
from typing import List

import langchain
import streamlit as st
import torch
import yaml
from dotenv import load_dotenv
from loguru import logger
from streamlit import chat_message

from llmsearch.config import Config
from llmsearch.chroma import VectorStoreChroma
from llmsearch.process import get_and_parse_response
from llmsearch.utils import get_llm_bundle, set_cache_folder
from llmsearch.embeddings import (
    update_embeddings,
    create_embeddings,
    EmbeddingsHashNotExistError,
)

st.set_page_config(page_title="LLMSearch", page_icon=":robot:", layout="wide")

load_dotenv()

langchain.debug = True


@st.cache_data
def parse_cli_arguments():
    parser = argparse.ArgumentParser(description="Web application for LLMSearch")
    parser.add_argument(
        "--doc_config_path", dest="cli_doc_config_path", type=str, default=""
    )
    parser.add_argument(
        "--model_config_path", dest="cli_model_config_path", type=str, default=""
    )
    try:
        args = parser.parse_args()
    except SystemExit as e:
        os._exit(e.code)
    return args


def hash_func(obj: Config) -> str:
    return str(obj.embeddings.embeddings_path)


def generate_index(config: Config):
    with st.spinner("Creading index, please wait..."):
        logger.debug("Unloading existing models...")
        unload_model()
        set_cache_folder(str(config.cache_folder))

        vs = VectorStoreChroma(
            persist_folder=str(config.embeddings.embeddings_path), config=config
        )
        create_embeddings(config, vs)
        logger.debug("Cleaning memory and re-Loading model...")
        vs.unload()
        vs = None
        gc.collect()
        with torch.no_grad():
            torch.cuda.empty_cache()

    st.success("Done generating index.")


def udpate_index(doc_config_path: str, model_config_file):
    """Updates index on-fly

    Args:
        config_file (str): _description_
    """

    with st.spinner("Updating index, please wait..."):
        logger.debug("Unloading model...")
        unload_model()
        config = load_config(doc_config_path, model_config_file)
        set_cache_folder(str(config.cache_folder))

        vs = VectorStoreChroma(
            persist_folder=str(config.embeddings.embeddings_path), config=config
        )
        try:
            logger.debug("Updating embeddings")
            stats = update_embeddings(config, vs)
        except EmbeddingsHashNotExistError:
            st.error(
                "Couldn't find hash files. Please re-create the index using current version of the app."
            )
        else:
            logger.info(stats)
        finally:
            logger.debug("Cleaning memory and re-Loading model...")

            vs.unload()

            vs = None

            gc.collect()
            with torch.no_grad():
                torch.cuda.empty_cache()

            reload_model(
                doc_config_path=doc_config_path, model_config_file=model_config_file
            )
    st.success("Done updating.")


@st.cache_data
def load_config(doc_config, model_config) -> Config:
    """Loads doc and model configurations, combines, and returns an instance of Config"""

    doc_config_dict = load_yaml_file(doc_config)
    model_config_dict = load_yaml_file(model_config)

    config_dict = {**doc_config_dict, **model_config_dict}
    return Config(**config_dict)


@st.cache_data
def load_yaml_file(config) -> dict:
    if isinstance(config, str):
        logger.info(f"Loading doc config from a file: {config}")
        with open(config, "r") as f:
            string_data = f.read()
    else:
        stringio = StringIO(config.getvalue().decode("utf-8"))
        string_data = stringio.read()

    config_dict = yaml.safe_load(string_data)
    return config_dict


def unload_model():
    """Unloads llm_bundle from the state to free up the GPU memory"""

    if st.session_state["llm_bundle"] is not None:
        st.session_state["llm_bundle"].store = None
        st.session_state["llm_bundle"].chain = None
        st.session_state["llm_bundle"].reranker = None
        st.session_state["llm_bundle"].hyde_chain = None
        st.session_state["llm_bundle"].multiquery_chain = None

    st.cache_data.clear()
    st.cache_resource.clear()
    gc.collect()

    st.session_state["llm_bundle"] = None
    gc.collect()

    with torch.no_grad():
        torch.cuda.empty_cache()


@st.cache_data
def generate_response(
    question: str,
    use_hyde: bool,
    use_multiquery,
    _config: Config,
    _bundle,
    label_filter: str = "",
):
    # _config and _bundle are under scored so paratemeters aren't hashed

    output = get_and_parse_response(
        query=question, config=_config, llm_bundle=_bundle, label=label_filter
    )
    return output


@st.cache_data
def get_config_paths(config_dir: str) -> List[str]:
    root = Path(config_dir)
    config_paths = sorted([str(p) for p in root.glob("*.yaml")])
    return config_paths


def reload_model(doc_config_path: str, model_config_file: str):
    if st.session_state["disable_load"]:
        logger.info("In process of loading the model, please wait...")
        return
    st.session_state["disable_load"] = True
    # This is required for memory management, we need to try and unload the model before loading new one

    logger.info("Clearing state and re-loading model...")
    unload_model()

    logger.debug(f"Reload model got DOC CONFIG FILE NAME: {doc_config_path}")
    logger.debug(f"Reload model got MODEL CONFIG FILE NAME: {model_config_file}")
    with st.spinner("Loading configuration"):
        config = load_config(doc_config_path, model_config_file)
        if config.check_embeddings_exist():
            st.session_state["llm_bundle"] = get_llm_bundle(config)
            st.session_state["llm_config"] = {
                "config": config,
                "doc_config_path": doc_config_path,
                "model_config_file": model_config_file,
            }
        else:
            st.error(
                "Couldn't find embeddings in {}. Please generate first.".format(
                    config.embeddings.embeddings_path
                )
            )
            update_embeddings_button = st.button(
                "Generate", on_click=generate_index, args=(config,), type="secondary"
            )

    st.session_state["disable_load"] = False


st.title(":sparkles: LLMSearch")
args = parse_cli_arguments()
st.sidebar.subheader(":hammer_and_wrench: Configuration")

# Initialsize state for historical resutls
if "llm_bundle" not in st.session_state:
    st.session_state["llm_bundle"] = None

if "llm_config" not in st.session_state:
    st.session_state["llm_config"] = None

# Initialsize state for historical resutls
if "messages" not in st.session_state:
    st.session_state["messages"] = []

if "disable_load" not in st.session_state:
    st.session_state["disable_load"] = False

if Path(args.cli_doc_config_path).is_dir():
    config_paths = get_config_paths(args.cli_doc_config_path)
    doc_config_path = st.sidebar.selectbox(
        label="Choose config", options=config_paths, index=0
    )
    model_config_file = args.cli_model_config_path
    logger.debug(f"CONFIG FILE: {doc_config_path}")

    # Every form must have a submit button.
    load_button = st.sidebar.button(
        "Load",
        on_click=reload_model,
        args=(doc_config_path, model_config_file),
        type="primary",
    )

    # Since in the event loop on_click will be called first, we need to re-enable the flag in case of multiple clicks
    if load_button:
        st.session_state["disable_load"] = False


if st.session_state["llm_bundle"] is not None:
    config = st.session_state["llm_config"]["config"]

    doc_config_path = st.session_state["llm_config"]["doc_config_path"]
    model_config_file = st.session_state["llm_config"]["model_config_file"]
    config_file_name = (
        doc_config_path if isinstance(doc_config_path, str) else doc_config_path.name
    )
    st.sidebar.subheader("Loaded Parameters:")
    with st.sidebar.expander(config_file_name):
        st.json(config.model_dump_json())

    st.sidebar.write(f"**Model type:** {config.llm.type}")

    st.sidebar.write(
        f"**Document path**: {config.embeddings.document_settings[0].doc_path}"
    )
    st.sidebar.write(f"**Embedding path:** {config.embeddings.embeddings_path}")
    update_embeddings_button = st.sidebar.button(
        "Update embeddings",
        on_click=udpate_index,
        args=(doc_config_path, model_config_file),
        type="secondary",
    )
    st.sidebar.write(
        f"**Max char size (semantic search):** {config.semantic_search.max_char_size}"
    )
    label_filter = ""
    if config.embeddings.labels:
        label_filter = st.sidebar.selectbox(
            label="Filter by label", options=["-"] + config.embeddings.labels
        )
        if label_filter is None or label_filter == "-":
            label_filter = ""

    text = st.chat_input("Enter text", disabled=False)
    is_hyde = st.sidebar.checkbox(
        label="Use HyDE (cost: 2 api calls)",
        value=st.session_state["llm_bundle"].hyde_enabled,
    )
    is_multiquery = st.sidebar.checkbox(
        label="Use MultiQuery (cost: 2 api calls)",
        value=st.session_state["llm_bundle"].multiquery_enabled,
    )

    if text:
        # Dynamically switch hyde
        st.session_state["llm_bundle"].hyde_enabled = is_hyde
        st.session_state["llm_bundle"].multiquery_enabled = is_multiquery
        output = generate_response(
            question=text,
            use_hyde=st.session_state["llm_bundle"].hyde_enabled,
            use_multiquery=st.session_state["llm_bundle"].multiquery_enabled,
            _bundle=st.session_state["llm_bundle"],
            _config=config,
            label_filter=label_filter,
        )

        # Add assistant response to chat history
        st.session_state["messages"].append(
            {
                "question": text,
                "response": output.response,
                "links": [
                    f'<a href="{s.chunk_link}">{s.chunk_link}</a>'
                    for s in output.semantic_search[::-1]
                ],
                "quality": f"{output.average_score:.2f}",
            }
        )
        for h_response in st.session_state["messages"]:
            with st.expander(
                label=f":question: **{h_response['question']}**", expanded=False
            ):
                st.markdown(f"##### {h_response['question']}")
                st.write(h_response["response"])
                st.markdown(
                    f"\n---\n##### Serrch Quality Score: {h_response['quality']}"
                )
                st.markdown("##### Links")
                for link in h_response["links"]:
                    st.write("\t* " + link, unsafe_allow_html=True)

        for source in output.semantic_search[::-1]:
            source_path = source.metadata.pop("source")
            score = source.metadata.get("score", None)
            with st.expander(label=f":file_folder: {source_path}", expanded=True):
                st.write(
                    f'<a href="{source.chunk_link}">{source.chunk_link}</a>',
                    unsafe_allow_html=True,
                )
                if score is not None:
                    st.write(f"\nScore: {score}")

                st.text(f"\n\n{source.chunk_text}")
        if st.session_state["llm_bundle"].hyde_enabled:
            with st.expander(label=":octagonal_sign: **HyDE Reponse**", expanded=False):
                st.write(output.hyde_response)

        with chat_message("assistant"):
            st.write(f"**Search results quality score: {output.average_score:.2f}**\n")
            st.write(output.response)  # Add user message to chat history


else:
    st.info("Choose a configuration template to start...")



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from collections import namedtuple


DBSettings = namedtuple("DBSettings", "SessionLocal engine")


def get_local_session(db_path: str) -> DBSettings:
    sql_alchemy_database_url = f"sqlite:///{db_path}"
    engine = create_engine(
        sql_alchemy_database_url, connect_args={"check_same_thread": False}
    )
    SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)
    return DBSettings(SessionLocal=SessionLocal, engine=engine)


Base = declarative_base()



import re
from collections import namedtuple
from enum import Enum
from pathlib import Path
from typing import Generator, List, Union, Tuple
from loguru import logger
import urllib

FORMATTING_SEQUENCES = {"*", "**", "***", "_", "__", "~~", "||"}
CODE_BLOCK_SEQUENCES = {"`", "``", "```"}
ALL_SEQUENCES = FORMATTING_SEQUENCES | CODE_BLOCK_SEQUENCES
MAX_FORMATTING_SEQUENCE_LENGTH = max(len(seq) for seq in ALL_SEQUENCES)


class SplitCandidates(Enum):
    SPACE = 1
    NEWLINE = 2
    LAST_CHAR = 3


# Order of preference for splitting
SPLIT_CANDIDATES_PREFRENCE = [
    SplitCandidates.NEWLINE,
    SplitCandidates.SPACE,
    SplitCandidates.LAST_CHAR,
]


BLOCK_SPLIT_CANDIDATES = [r"\n#\s+", r"\n##\s+", r"\n###\s+"]
CODE_BLOCK_LEVEL = 10  # SHould be high enough to rank it above everything else

MarkdownChunk = namedtuple("MarkdownChunk", "string level")


class SplitCandidateInfo:
    last_seen: int
    active_sequences: List[str]
    active_sequences_length: int

    def __init__(self):
        self.last_seen = None
        self.active_sequences = []
        self.active_sequences_length = 0

    def process_sequence(self, seq: str, is_in_code_block: bool):
        """Process `seq`, update `self.active_sequences` and `self.active_sequences_length`,
        and return whether we are in a code block after processing `seq`.
        """

        if is_in_code_block:
            if self.active_sequences and seq == self.active_sequences[-1]:
                last_seq = self.active_sequences.pop()
                self.active_sequences_length -= len(last_seq)
            return True
        elif seq in CODE_BLOCK_SEQUENCES:
            self.active_sequences.append(seq)
            self.active_sequences_length += len(seq)
            return True
        else:
            for k in range(len(self.active_sequences) - 1, -1, -1):
                if seq == self.active_sequences[k]:
                    sequences_being_removed = self.active_sequences[k:]
                    self.active_sequences = self.active_sequences[:k]
                    self.active_sequences_length -= sum(
                        len(seq) for seq in sequences_being_removed
                    )
                    return False
            self.active_sequences.append(seq)
            self.active_sequences_length += len(seq)
            return False

    def copy_from(self, other):
        self.last_seen = other.last_seen
        self.active_sequences = other.active_sequences.copy()
        self.active_sequences_length = other.active_sequences_length


def phsyical_split(markdown: str, max_chunk_size: int) -> Generator[str, None, None]:
    """Naive markdown splitter that splits long messages in chunks
    preserving the markdown formatting tags. This split method isn't aware of higher level logical blocks,
    but preserves the low level blocks
    """

    if max_chunk_size <= MAX_FORMATTING_SEQUENCE_LENGTH:
        raise ValueError(
            f"max_chunk_size must be greater than {MAX_FORMATTING_SEQUENCE_LENGTH}"
        )

    split_candidates = {
        SplitCandidates.SPACE: SplitCandidateInfo(),
        SplitCandidates.NEWLINE: SplitCandidateInfo(),
        SplitCandidates.LAST_CHAR: SplitCandidateInfo(),
    }
    is_in_code_block = False

    chunk_start_from, chunk_char_count, chunk_prefix = 0, 0, ""

    def split_chunk():
        for split_variant in SPLIT_CANDIDATES_PREFRENCE:
            split_candidate = split_candidates[split_variant]
            if split_candidate.last_seen is None:
                continue
            chunk_end = split_candidate.last_seen + (
                1 if split_variant == SplitCandidates.LAST_CHAR else 0
            )
            chunk = (
                chunk_prefix
                + markdown[chunk_start_from:chunk_end]
                + "".join(reversed(split_candidate.active_sequences))
            )

            next_chunk_prefix = "".join(split_candidate.active_sequences)
            next_chunk_char_count = len(next_chunk_prefix)
            next_chunk_start_from = chunk_end + (
                0 if split_variant == SplitCandidates.LAST_CHAR else 1
            )

            split_candidates[SplitCandidates.NEWLINE] = SplitCandidateInfo()
            split_candidates[SplitCandidates.SPACE] = SplitCandidateInfo()
            return (
                chunk,
                next_chunk_start_from,
                next_chunk_char_count,
                next_chunk_prefix,
            )

    i = 0
    while i < len(markdown):
        for j in range(MAX_FORMATTING_SEQUENCE_LENGTH, 0, -1):
            seq = markdown[i : i + j]
            if seq in ALL_SEQUENCES:
                last_char_split_candidate_len = (
                    chunk_char_count
                    + split_candidates[
                        SplitCandidates.LAST_CHAR
                    ].active_sequences_length
                    + len(seq)
                )
                if last_char_split_candidate_len >= max_chunk_size:
                    (
                        next_chunk,
                        chunk_start_from,
                        chunk_char_count,
                        chunk_prefix,
                    ) = split_chunk()
                    yield next_chunk
                is_in_code_block = split_candidates[
                    SplitCandidates.LAST_CHAR
                ].process_sequence(seq, is_in_code_block)
                i += len(seq)
                chunk_char_count += len(seq)
                split_candidates[SplitCandidates.LAST_CHAR].last_seen = i - 1
                break

        if i >= len(markdown):
            break

        split_candidates[SplitCandidates.LAST_CHAR].last_seen = i
        chunk_char_count += 1
        if markdown[i] == "\n":
            split_candidates[SplitCandidates.NEWLINE].copy_from(
                split_candidates[SplitCandidates.LAST_CHAR]
            )
        elif markdown[i] == " ":
            split_candidates[SplitCandidates.SPACE].copy_from(
                split_candidates[SplitCandidates.LAST_CHAR]
            )

        last_char_split_candidate_len = (
            chunk_char_count
            + split_candidates[SplitCandidates.LAST_CHAR].active_sequences_length
        )
        if last_char_split_candidate_len == max_chunk_size:
            next_chunk, chunk_start_from, chunk_char_count, chunk_prefix = split_chunk()
            yield next_chunk

        i += 1

    if chunk_start_from < len(markdown):
        yield chunk_prefix + markdown[chunk_start_from:]


def get_logical_blocks_recursively(
    markdown: str, max_chunk_size: int, all_sections: list, split_candidate_index=0
) -> List[MarkdownChunk]:
    """Recursively scans blocks, splittling the larger blocks using next available paragraph size

    Args:
        markdown (str): Markdown to split
        max_chunk_size (int): Maximum chunk size
        all_sections (list): Keeps tracks of all sections
        split_candidate_index (int, optional): Index of split sequence in BLOCK_SPLIT_CANDIDATEs. Starts from 0.

    Returns:
        List[str]: List of logically split chunks
    """

    if split_candidate_index >= len(BLOCK_SPLIT_CANDIDATES):
        for chunk in phsyical_split(markdown, max_chunk_size):
            all_sections.append(
                MarkdownChunk(string=chunk, level=split_candidate_index)
            )
        return all_sections
    # else:
    #     split_candidate = BLOCK_SPLIT_CANDIDATES[split_candidate_index]

    chunks = []
    add_index = 0
    for add_index, split_candidate in enumerate(
        BLOCK_SPLIT_CANDIDATES[split_candidate_index:]
    ):
        chunks = re.split(split_candidate, markdown)
        if len(chunks) > 1:
            break

    for i, chunk in enumerate(chunks):
        level = split_candidate_index + add_index

        # First chunk is left-over from a previous chunk
        if i > 0:
            level += 1

        prefix = "\n\n" + "#" * level + " "
        if not chunk.strip():
            continue

        if len(chunk) <= max_chunk_size:
            all_sections.append(MarkdownChunk(string=prefix + chunk, level=level - 1))
        else:
            get_logical_blocks_recursively(
                chunk,
                max_chunk_size,
                all_sections,
                split_candidate_index=split_candidate_index + add_index + 1,
            )
    return all_sections


def markdown_splitter(
    path: Union[str, Path], max_chunk_size: int, **additional_splitter_settings
) -> List[dict]:
    """Logical split based on top-level headings.

    Args:
        markdown (str): markdown string
        max_chunk_size (int): Maximum chunk size
    """

    try:
        with open(path, "r") as f:
            markdown = f.read()
    except OSError:
        return []

    if len(markdown) < max_chunk_size:
        return [{"text": markdown, "metadata": {"heading": ""}}]

    sections = [MarkdownChunk(string="", level=0)]

    markdown, additional_metadata = preprocess_markdown(
        markdown, additional_splitter_settings
    )

    # Split by code and non-code
    chunks = markdown.split("```")

    for i, chunk in enumerate(chunks):
        if i % 2 == 0:  # Every even element (0 indexed) is a non-code
            logical_blocks = get_logical_blocks_recursively(
                chunk, max_chunk_size=max_chunk_size, all_sections=[]
            )
            sections += logical_blocks
        else:  # Process the code section
            rows = chunk.split("\n")
            code = rows[1:]

            lang = rows[0]  # Get the language name

            # Provide a hint to LLM
            all_code_rows = (
                [
                    f"\nFollowing is a code section in {lang}, delimited by triple backticks:",
                    f"```{lang}",
                ]
                + code
                + ["```"]
            )
            all_code_str = "\n".join(all_code_rows)

            # Merge code to a previous logical block if there is enough space
            if len(sections[-1].string) + len(all_code_str) < max_chunk_size:
                sections[-1] = MarkdownChunk(
                    string=sections[-1].string + all_code_str, level=sections[-1].level
                )

            # If code block is larger than max size, physically split it
            elif len(all_code_str) >= max_chunk_size:
                code_chunks = phsyical_split(
                    all_code_str, max_chunk_size=max_chunk_size
                )
                for cchunk in code_chunks:
                    # Assign language header to the code chunk, if doesn't exist
                    if f"```{lang}" not in cchunk:
                        cchunk_rows = cchunk.split("```")
                        cchunk = f"```{lang}\n" + cchunk_rows[1] + "```"

                    sections.append(
                        MarkdownChunk(string=cchunk, level=CODE_BLOCK_LEVEL)
                    )

            # Otherwise, add as a single chunk
            else:
                sections.append(
                    MarkdownChunk(string=all_code_str, level=CODE_BLOCK_LEVEL)
                )

    all_out = postprocess_sections(
        sections,
        max_chunk_size,
        additional_splitter_settings,
        additional_metadata,
        path,
    )

    logger.info(f"Got {len(all_out)} text chunks:")
    for s in all_out:
        logger.info(f"\tChunk length: {len(s['text'])}")
    return all_out


def preprocess_markdown(markdown: str, additional_settings: dict) -> Tuple[str, dict]:
    """Perform pre-processing, before splitting the document into logical blocks

    Args:
        markdown (str): string containing the document
        additional_settings (dict): Dictionary of additional pre-processing settings:
                                    remove_images (bool) - removes image links the document
                                    remove_extra_newlines (bool) - removes extra new lines (3 or more)

    Returns:
        str, dict: processed markdown string, dictionary with additional metadata
    """

    preprocess_remove_images = additional_settings.get("remove_images", False)
    preprocess_remove_extra_newlines = additional_settings.get(
        "remove_extra_newlines", True
    )
    preprocess_find_metadata = additional_settings.get("find_metadata", dict())

    if preprocess_remove_images:
        markdown = remove_images(markdown)

    if preprocess_remove_extra_newlines:
        markdown = remove_extra_newlines(markdown)

    additional_metadata = {}

    if preprocess_find_metadata:
        if not isinstance(preprocess_find_metadata, dict):
            raise TypeError(
                f"find_metadata settings should be of type dict. Got {type(preprocess_find_metadata)}"
            )

        for label, search_string in preprocess_find_metadata.items():
            logger.info(f"Looking for metadata: {search_string}")
            metadata = find_metadata(markdown, search_string)
            if metadata:
                logger.info(f"\tFound metadata for {label} - {metadata}")
                additional_metadata[label] = metadata

    return markdown, additional_metadata


def postprocess_sections(
    sections: List[MarkdownChunk],
    max_chunk_size: int,
    additional_settings: dict,
    additional_metadata: dict,
    path: Union[str, Path],
) -> List[dict]:
    all_out = []

    skip_first = additional_settings.get("skip_first", False)
    merge_headers = additional_settings.get("merge_sections", False)

    # Remove all empty sections
    sections = [s for s in sections if s.string]

    if sections and skip_first:
        logger.info("Removing first section ...")
        sections = sections[1:]

    if sections and merge_headers:
        logger.info("Merging sections ...")
        sections = merge_sections(sections, max_chunk_size=max_chunk_size)

    current_heading = ""

    sections_metadata = {"Document name": Path(path).name}

    for s in sections:
        stripped_string = s.string.strip()
        doc_metadata = {}
        if len(stripped_string) > 0:
            heading = ""
            if stripped_string.startswith("#"):  # heading detected
                heading = stripped_string.split("\n")[0].replace("#", "").strip()
                stripped_heading = heading.replace("#", "").replace(" ", "").strip()
                if not stripped_heading:
                    heading = ""
                if s.level == 0:
                    current_heading = heading
                doc_metadata["heading"] = urllib.parse.quote(
                    heading
                )  # isolate the heading
            else:
                doc_metadata["heading"] = ""

            final_section = add_section_metadata(
                stripped_string,
                section_metadata={
                    **sections_metadata,
                    **{"Subsection of": current_heading},
                    **additional_metadata,
                },
            )
            all_out.append({"text": final_section, "metadata": doc_metadata})
    return all_out


def remove_images(page_md: str) -> str:
    logger.info("Removing images ... ")
    return re.sub(r"""!\[[^\]]*\]\((.*?)\s*("(?:.*[^"])")?\s*\)""", "", page_md)


def remove_extra_newlines(page_md) -> str:
    """Remove extra new lines from the text if there are 3 or more consecutive new lines

    Args:
        page_md (str): Markdown page in a string fromate

    Returns:
        str: Markdown string with nre lines removed
    """

    logger.info("Removing extra new lines ... ")
    page_md = re.sub(r"\n{3,}", "\n\n", page_md)
    return page_md


def add_section_metadata(s, section_metadata: dict):
    metadata_s = ""
    for k, v in section_metadata.items():
        if v:
            metadata_s += f"{k}: {v}\n"
    metadata = f"Metadata applicable to the next chunk of text delimited by five stars:\n>> METADATA START\n{metadata_s}>> METADATA END\n\n"

    return metadata + "*****\n" + s + "\n*****"


def find_metadata(page_md: str, search_string: str) -> str:
    pattern = rf"{search_string}(.*)"
    match = re.search(pattern, page_md)
    if match:
        return match.group(1)
    return ""


def merge_sections(
    sections: List[MarkdownChunk], max_chunk_size: int
) -> List[MarkdownChunk]:
    """Merges sections together, up to a max chunk size"""

    current_section = sections[0]
    all_out = []

    prev_level = 0
    for s in sections[1:]:
        if (
            len(current_section.string + s.string) > max_chunk_size
            or s.level <= prev_level
        ):
            all_out.append(current_section)
            current_section = s
            prev_level = 0
        else:
            current_section = MarkdownChunk(
                string=current_section.string + s.string, level=current_section.level
            )
            prev_level = s.level if s.level != CODE_BLOCK_LEVEL else prev_level

    all_out.append(current_section)

    return all_out


if __name__ == "__main__":
    from pathlib import Path

    path = Path(
        "/storage/gdp-platform-docs/docs/platform/how-to/onboard/get-onboard-data-platform.md"
    )
    # path = Path("/storage/gdp-platform-docs/docs/platform/how-to/release/synapse_cicd_release.md")
    # path = Path("/storage/gdp-platform-docs/docs/platform/how-to/connect/get-started-azure-data-studio.md")

    print("**************************************************")
    # chunks = get_logical_blocks_recursively(text, all_sections = [], max_chunk_size=1024)
    chunks = markdown_splitter(
        path=path,
        max_chunk_size=1024,
        **{
            "merge_sections": True,
            "skip_first": True,
            "remove_images": True,
            "find_metadata": {"description": "description: "},
        },
    )
    print(len(chunks))

    for chunk in chunks:
        print("\n\nSTART CHUNK ----------------")
        print(chunk["text"])
        print("END CHUNK ----------------")

    for chunk in chunks:
        print(len(chunk["text"]))



import string
from typing import List

from loguru import logger
from llmsearch.config import (
    SuffixAppend,
    ObsidianAdvancedURI,
    ResponseModel,
    SemanticSearchOutput,
    Config,
)
from llmsearch.ranking import get_relevant_documents
from llmsearch.utils import LLMBundle
from llmsearch.database.crud import create_response


def get_and_parse_response(
    llm_bundle: LLMBundle,
    query: str,
    config: Config,
    persist_db_session=None,
    label: str = "",
) -> ResponseModel:
    """Performs retieval augmented search (RAG).

    Args:
        llm_bundle (LLMBundle): Runtime parameters for LLM and retrievers
        config (SemanticSearchConfig): Configuration

    Returns:
        OutputModel
    """

    original_query = query

    queries = []
    hyde_response = ""
    # Add HYDE queries, if required
    if llm_bundle.hyde_enabled:
        hyde_response = get_hyde_response(llm_bundle, query)
        queries += [hyde_response]

    elif llm_bundle.multiquery_enabled:
        queries += get_multiquery_response(
            llm_bundle, query, config.semantic_search.multiquery.n_versions
        )
    else:
        queries = [query]

    semantic_search_config = config.semantic_search
    most_relevant_docs, score = get_relevant_documents(
        original_query, queries, llm_bundle, semantic_search_config, label=label
    )

    res = llm_bundle.chain(
        {"input_documents": most_relevant_docs, "question": original_query},
        return_only_outputs=False,
    )

    out = ResponseModel(
        response=res["output_text"],
        question=query,
        average_score=score,
        hyde_response=hyde_response,
    )
    for doc in res["input_documents"]:
        doc_name = doc.metadata["source"]

        for replace_setting in semantic_search_config.replace_output_path:
            doc_name = doc_name.replace(
                replace_setting.substring_search,
                replace_setting.substring_replace,
            )

        if semantic_search_config.obsidian_advanced_uri is not None:
            doc_name = process_obsidian_uri(
                doc_name, semantic_search_config.obsidian_advanced_uri, doc.metadata
            )

        if semantic_search_config.append_suffix is not None:
            doc_name = process_append_suffix(
                doc_name, semantic_search_config.append_suffix, doc.metadata
            )

        text = doc.page_content
        out.semantic_search.append(
            SemanticSearchOutput(
                chunk_link=doc_name, metadata=doc.metadata, chunk_text=text
            )
        )

    if llm_bundle.response_persist_db_settings is not None:
        if persist_db_session is not None:
            session = persist_db_session
        else:
            session = llm_bundle.response_persist_db_settings.SessionLocal()
        create_response(config=config, session=session, response=out)

        # if locally created session, close it, otherwise it is assumed to be externally closed (e.g. api)
        if not persist_db_session:
            session.close()
    return out


def process_obsidian_uri(
    doc_name: str, adv_uri_config: ObsidianAdvancedURI, metadata: dict
) -> str:
    """Adds a suffix pointing to a specific heading based on the metadata supplied if doc.metadata

    Args:
        doc_name (str): Document name (partially processed, potentially)
        adv_uri_config (ObsidianAdvancedURI): contains the template to add,
                                              matches Obsidian's advanced URI plugin schem
        metadata (dict): Metadata associated with a document.

    Returns:
        str: document name with a header suffix.
    """
    print(metadata)
    append_str = adv_uri_config.append_heading_template.format(
        heading=metadata["heading"]
    )
    return doc_name + append_str


def process_append_suffix(doc_name, suffix: SuffixAppend, metadata: dict):
    fmt = PartialFormatter(missing="")
    return doc_name + fmt.format(suffix.append_template, **metadata)


class PartialFormatter(string.Formatter):
    def __init__(self, missing="~~", bad_fmt="!!"):
        self.missing, self.bad_fmt = missing, bad_fmt

    def get_field(self, field_name, args, kwargs):
        # Handle a key not found
        try:
            val = super(PartialFormatter, self).get_field(field_name, args, kwargs)
            # Python 3, 'super().get_field(field_name, args, kwargs)' works
        except (KeyError, AttributeError):
            val = None, field_name
        return val

    def format_field(self, value, spec):
        # handle an invalid format
        if value is None:
            return self.missing
        try:
            return super(PartialFormatter, self).format_field(value, spec)
        except ValueError:
            if self.bad_fmt is not None:
                return self.bad_fmt
            else:
                raise


def get_hyde_response(llm_bundle: LLMBundle, query: str) -> str:
    """Generates an artificial response using HyDE technique (Hypothetical Document Embeddings)
    See https://arxiv.org/pdf/2212.10496.pdf for more detail

    Args:
        llm_bundle (LLMBundle): instance of LLMBundle containing runtime parameters.
        query (str): Original query

    Returns:
        str: Artifical passage which answers the query
    """

    if llm_bundle.hyde_chain is None:
        raise TypeError("HyDE chain is not initialised. exiting.")
    res = llm_bundle.hyde_chain.run(query)
    logger.info(f"HYDE: got response: {res}")
    return res


def get_multiquery_response(
    llm_bundle: LLMBundle, query: str, n_versions: int
) -> List[str]:
    if llm_bundle.multiquery_chain is None:
        raise TypeError("MultiQuery chain is not initialised. exiting.")
    res = llm_bundle.multiquery_chain.run(question=query, n_versions=n_versions)

    logger.info(f"MultiQuery: got response: {res}")
    queries = [q.strip() for q in res.strip().split("\n") if q.strip()]
    if len(queries) != n_versions:
        raise ValueError(
            "Number of versions in multi-queries response isn't equal {}".format(
                n_versions
            )
        )
    return queries



import click
from loguru import logger
import os
import streamlit.web.cli

from llmsearch.chroma import VectorStoreChroma
from llmsearch.config import get_config, get_doc_with_model_config
from llmsearch.interact import qa_with_llm
from llmsearch.utils import get_llm_bundle, set_cache_folder
from llmsearch.embeddings import create_embeddings, update_embeddings


@click.group(name="index")
def index_group():
    """Index generation commands."""


@click.group(name="interact")
def interact_group():
    """Commands to interact in Q&A sessiont with embedded content using LLMs"""


@click.group
def main_cli():
    pass


@click.command(name="create")
@click.option(
    "--config-file",
    "-c",
    "config_file",
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help="Specifies YAML configuration file",
)
def generate_index(config_file: str):
    config = get_config(config_file)
    set_cache_folder(str(config.cache_folder))

    vs = VectorStoreChroma(
        persist_folder=str(config.embeddings.embeddings_path), config=config
    )
    create_embeddings(config, vs)


@click.command(name="update")
@click.option(
    "--config-file",
    "-c",
    "config_file",
    required=True,
    type=click.Path(exists=True, dir_okay=False),
    help="Specifies YAML configuration file",
)
def udpate_index(config_file: str):
    config = get_config(config_file)
    set_cache_folder(str(config.cache_folder))

    vs = VectorStoreChroma(
        persist_folder=str(config.embeddings.embeddings_path), config=config
    )
    stats = update_embeddings(config, vs)
    logger.info(stats)


@click.command("llm")
@click.option(
    "--doc-config-file",
    "-c",
    "doc_config_file",
    required=True,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    help="Specifies documents YAML configuration file",
)
@click.option(
    "--model-config-file",
    "-m",
    "model_config_file",
    required=True,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    help="Specifies model YAML configuration file",
)
def launch_qa_with_llm(doc_config_file: str, model_config_file: str):
    config = get_doc_with_model_config(doc_config_file, model_config_file)
    llm_bundle = get_llm_bundle(config)
    qa_with_llm(llm_bundle, config)


@click.command("webapp")
@click.option(
    "--doc-config-path",
    "-c",
    "doc_config_path",
    required=True,
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="Specifies documents YAML configuration file",
)
@click.option(
    "--model-config-file",
    "-m",
    "model_config_file",
    required=True,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    help="Specifies model YAML configuration file",
)
def launch_streamlit(doc_config_path: str, model_config_file: str):
    # Based on
    # https://discuss.streamlit.io/t/running-streamlit-inside-my-own-executable-with-the-click-module/1198/4
    # streamlit run ./src/llmsearch/webapp.py -- --config_path ./sample_templates/obsidian_conf.yaml

    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, "webapp.py")
    args = [
        "--doc_config_path",
        doc_config_path,
        "--model_config_path",
        model_config_file,
    ]
    streamlit.web.cli._main_run(filename, args)


index_group.add_command(generate_index)
index_group.add_command(udpate_index)
interact_group.add_command(launch_qa_with_llm)

index_group.add_command(generate_index)
interact_group.add_command(launch_qa_with_llm)
interact_group.add_command(launch_streamlit)

# add command groups to CLI root
main_cli.add_command(index_group)
main_cli.add_command(interact_group)

if __name__ == "__main__":
    main_cli()



DOLLY_PROMPT_TEMPLATE = """### Instruction:
Use the following pieces of context to answer the question at the end. If answer isn't in the context, say that you don't know, don't try to make up an answer.

### Context: 
---------------
{context}
---------------

### Question: {question}
"""


WIZARDLM10_TEMPLATE = """
A chat between a curious user and an artificial intelligence assistant. The assistant gives an answer to the user's questions, only if it exists in the provided context, and refuses to answer otherwise.

CONTEXT:
---------------
{context}
---------------

USER: {question}
ASSISTANT:
"""

LLAMA_TEMPLATE = """### Instruction:
Use the following pieces of context to answer the question at the end. If answer isn't in the context, say that you don't know, don't try to make up an answer.

### Context: 
---------------
{context}
---------------

### Question: {question}
### Response:"""


NOUS_HERMES_TEMPLATE = """### Instruction:
Use the context provided below to answer user request at the end, using only information in the context. If answer isn't in the context, say that you don't know, don't try to make up an answer.

### Context: 
---------------
{context}
---------------

### User: {question}
### Response:"""


OPENAI_PROMPT_TEMPLATE = """"Context information is provided below. Given the context information and not prior knowledge, answer the question. If answer isn't in the context, say that you don't know, don't try to make up an answer.

### Context: 
---------------------
{context}
---------------------

### Question: {question}
"""

TULU8_TEMPLATE = """### Human: Use information in the provided context to answer the following question: ```{question}```.

### Context: 
---------------
{context}
---------------

### Assistant:
"""


REDPAJAMA_TEMPLATE = """Given the context and no prior information, answer the question.

# Context:
------------
{context}
------------
# Question: {question}
# Answer:
"""



from unstructured.partition.pdf import partition_pdf

import fitz

if __name__ == "__main__":
    n_output_elements = 2000
    path = "/storage/llm/pdf_docs/Patrick Viafore - Robust Python_ Write Clean and Maintainable Code-O'Reilly Media (2021).pdf"
    doc = fitz.open(path)
    with open("output_mupdf.txt", mode="w") as f:
        for i, page in enumerate(doc):
            block = page.get_text("block")
            f.write(block)
            f.write("**********\n")
        # if i == 100:
        #     break
        # print("-----------")
        # print(block)
        # print("-----------")
        # print(page.get_text("block"))

    elements = partition_pdf(
        filename=path,
        strategy="fast",
        include_page_breaks=True,
        infer_table_structure=True,
    )
    print(elements)
    print("Done processing. Writing output...")

    with open("output.txt", mode="w") as f:
        for el in elements:  # [:n_output_elements]:
            # s = f"<{el.category} - {el.metadata.page_number}> {el.text}\n"
            s = f"{el.text}\n"
            f.write(s)



import hashlib
import uuid

from loguru import logger
from sqlalchemy.orm import Session

from llmsearch.config import Config, ResponseModel
from llmsearch.database import models


class ResponseInteractionLookupError(Exception):
    """Raised in case response with a given response id doesn't exist"""


def get_or_store_config(config: Config, session: Session) -> str:
    config_json = config.json()
    config_hash = hashlib.md5(config_json.encode()).hexdigest()
    print(config_hash)

    config_id = (
        session.query(models.Config.config_id)
        .filter(models.Config.config_hash == config_hash)
        .scalar()
    )
    if config_id is None:
        logger.warning("Config doesn't exist in the database, creating new entry...")

        config_id = str(uuid.uuid4())

        db_config = models.Config(
            config_id=config_id, config_hash=config_hash, config_json=config_json
        )
        session.add(db_config)
        session.commit()
    logger.info(f"Retrieved config id: {config_id}")
    return config_id


def create_response(
    config: Config, session: Session, response: ResponseModel
) -> models.ResponseInteraction:
    config_id = get_or_store_config(config, session)

    sources = []
    for ss_source in response.semantic_search:
        sources.append(
            models.Sources(
                source_id=models.create_uuid(),
                text_blob=ss_source.chunk_text,
                text_link=ss_source.chunk_link,
                rank_score=ss_source.metadata["score"],
                additional_metadata=ss_source.metadata,
            )
        )

    db_response_interaction = models.ResponseInteraction(
        response_id=response.id,
        question_text=response.question,
        response_text=response.response,
        average_score=response.average_score,
        config_id=config_id,
    )

    # Add bridge records
    bridges = []
    for s in sources:
        session.add(s)
        session.add(
            models.InteractionSourcesBridge(
                response_id=db_response_interaction.response_id, source_id=s.source_id
            )
        )

    session.add(db_response_interaction)
    session.add_all(bridges)
    session.add_all(sources)
    session.commit()
    logger.info(
        f"Saved response to the database. Response id: {db_response_interaction.response_id}"
    )
    return db_response_interaction


def create_feedback(
    session: Session, response_id: str, is_positive: bool, feedback_text: str = ""
):
    logger.info(
        f"Updating response {response_id} with values: is_positive = {is_positive}, feedback_text = '{feedback_text}'"
    )
    resp_int = session.query(models.ResponseInteraction).filter(
        models.ResponseInteraction.response_id == response_id
    )
    if resp_int.scalar() is None:
        raise ResponseInteractionLookupError(
            f"Response id {response_id} doesn't exist."
        )

    resp_int.update({"is_positive": is_positive, "feedback_text": feedback_text})
    session.commit()



from typing import Union

from llmsearch.models.config import (
    OpenAIModelConfig,
    LlamaModelConfig,
    #    AutoGPTQModelConfig,
    HuggingFaceModelConfig,
    AzureOpenAIModelConfig,
)

from llmsearch.models.llama import LlamaModel
from llmsearch.models.hf import HuggingFaceModel
from llmsearch.models.openai import OpenAIModel
from llmsearch.models.azureopenai import AzureOpenAIModel

# from llmsearch.models.autogptq import AutoGPTQModel

model_mappings = {
    LlamaModelConfig: LlamaModel,
    HuggingFaceModelConfig: HuggingFaceModel,
    OpenAIModelConfig: OpenAIModel,
    AzureOpenAIModelConfig: AzureOpenAIModel,
    #  AutoGPTQModelConfig: AutoGPTQModel,
}


def get_llm(
    llm_config: Union[
        OpenAIModelConfig,
        LlamaModelConfig,
        # AutoGPTQModelConfig,
        HuggingFaceModelConfig,
    ]
):
    model_type = model_mappings.get(type(llm_config), None)  # type: ignore
    if model_type is None:
        raise TypeError(f"Unknown model type: {type(llm_config)}")

    llm = model_type(llm_config)  # type: ignore
    return llm



from llmsearch.embeddings import get_changed_or_new_files
import pandas as pd


def test_new_only():
    new_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
            {"filehash": "33333", "filename": "file3"},
        ]
    )
    existing_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
        ]
    )

    changed_or_new, changed, deleted = get_changed_or_new_files(
        new_hashes, existing_hashes
    )
    assert len(changed_or_new) == 1
    assert len(changed) == 0
    assert len(deleted) == 0

    assert changed_or_new.iloc[0, 1] == "file3"


def test_delete_only():
    new_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
        ]
    )
    existing_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
            {"filehash": "33333", "filename": "file3"},
        ]
    )

    changed_or_new, changed, deleted = get_changed_or_new_files(
        new_hashes, existing_hashes
    )
    assert len(changed_or_new) == 0
    assert len(changed) == 0
    assert len(deleted) == 1

    assert deleted.iloc[0, 1] == "file3"


def test_change_only():
    new_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
            {"filehash": "33331", "filename": "file3"},
            {"filehash": "44441", "filename": "file4"},
        ]
    )
    existing_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
            {"filehash": "33333", "filename": "file3"},
            {"filehash": "44444", "filename": "file4"},
        ]
    )

    changed_or_new, changed, deleted = get_changed_or_new_files(
        new_hashes, existing_hashes
    )
    assert len(changed_or_new) == 2
    assert len(changed) == 2
    assert len(deleted) == 0

    assert changed_or_new.iloc[0, 1] == "file3"


def test_new_and_delete():
    new_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
            {"filehash": "33333", "filename": "file3"},
            {"filehash": "55555", "filename": "file5"},
        ]
    )
    existing_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
            {"filehash": "33333", "filename": "file3"},
            {"filehash": "44444", "filename": "file4"},
        ]
    )

    changed_or_new, changed, deleted = get_changed_or_new_files(
        new_hashes, existing_hashes
    )
    assert len(changed_or_new) == 1
    assert len(changed) == 0
    assert len(deleted) == 1

    assert changed_or_new.iloc[0, 1] == "file5"


def test_new_change_and_delete():
    new_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
            {"filehash": "33333", "filename": "file3"},
            {"filehash": "55555", "filename": "file5"},
        ]
    )
    existing_hashes = pd.DataFrame(
        [
            {"filehash": "11111", "filename": "file1"},
            {"filehash": "22222", "filename": "file2"},
            {"filehash": "33332", "filename": "file3"},
            {"filehash": "44444", "filename": "file4"},
        ]
    )

    changed_or_new, changed, deleted = get_changed_or_new_files(
        new_hashes, existing_hashes
    )
    assert len(changed_or_new) == 2
    assert len(changed) == 1
    assert len(deleted) == 1

    assert changed.iloc[0, 1] == "file3"
    assert deleted.iloc[0, 1] == "file4"



import torch
import transformers
from langchain_community.llms import HuggingFacePipeline

from llmsearch.models.abstract import AbstractLLMModel
from llmsearch.models.config import HuggingFaceModelConfig
from transformers import BitsAndBytesConfig


class HuggingFaceModel(AbstractLLMModel):
    def __init__(self, config: HuggingFaceModelConfig) -> None:
        super().__init__(prompt_template=config.prompt_template)
        self.config = config

    @property
    def model(self):

        bitsandbytesconf = BitsAndBytesConfig(
            load_in_8bit=self.config.load_8bit, llm_int8_enable_fp32_cpu_offload=True
        )

        tokenizer_name = (
            self.config.tokenizer_name
            if self.config.tokenizer_name is not None
            else self.config.model_name
        )
        tokenizer = transformers.AutoTokenizer.from_pretrained(
            tokenizer_name, **self.config.tokenzier_kwargs
        )
        # model_kwargs = self.config.model_kwargs.update({"cache_dir": self.config.cache_folder})

        model_ = transformers.AutoModelForCausalLM.from_pretrained(
            self.config.model_name,
            cache_dir=self.config.cache_folder,
            device_map="auto",
            trust_remote_code=self.config.trust_remote_code,
            quantization_config=bitsandbytesconf,
            # load_in_8bit=self.config.load_8bit,
            **self.config.model_kwargs
        )

        pipeline = transformers.pipeline(
            "text-generation",
            model=model_,
            tokenizer=tokenizer,
            eos_token_id=tokenizer.eos_token_id,
            model_kwargs=self.config.model_kwargs,
            **self.config.pipeline_kwargs
        )

        llm = HuggingFacePipeline(pipeline=pipeline)
        return llm



from collections import defaultdict
import os
import pickle
from typing import List, Tuple

import numpy as np
import scipy
import torch
import tqdm
from loguru import logger
from transformers import AutoModelForMaskedLM, AutoTokenizer

# from llmsearch.utils import set_cache_folder
from llmsearch.config import Config, Document


def split(iterable: List, chunk_size: int):
    """Splits a list to chunks of size `chunk_size`"""

    for i in range(0, len(iterable), chunk_size):
        yield iterable[i : i + chunk_size]


class SparseEmbeddingsSplade:
    def __init__(
        self,
        config: Config,
        splade_model_id: str = "naver/splade-cocondenser-ensembledistil",
    ) -> None:
        self._config = config

        self._device = (
            f"cuda:{torch.cuda.current_device()}"
            if torch.cuda.is_available()
            else "cpu"
        )
        logger.info(f"Setting device to {self._device}")

        #        set_cache_folder(str(config.cache_folder))
        self.tokenizer = AutoTokenizer.from_pretrained(
            splade_model_id, device=self._device, use_fast=True
        )
        self.model = AutoModelForMaskedLM.from_pretrained(splade_model_id)
        self.model.to(self._device)
        self._embeddings = None
        self._ids = None
        self._l2_norm_matrix = None
        self._labels_to_ind = defaultdict(list)
        self._chunk_size_to_ind = defaultdict(list)

        self.n_batch = config.embeddings.splade_config.n_batch

    def _get_batch_embeddings(
        self, docs: List[str], free_memory: bool = True
    ) -> np.ndarray:
        tokens = self.tokenizer(
            docs, return_tensors="pt", padding=True, truncation=True
        ).to(self._device)

        output = self.model(**tokens)

        # aggregate the token-level vecs and transform to sparse
        vecs = (
            torch.max(
                torch.log(1 + torch.relu(output.logits))
                * tokens.attention_mask.unsqueeze(-1),
                dim=1,
            )[0]
            .squeeze()
            .detach()
            .cpu()
            .numpy()
        )

        # For smaller VRAM sizes, might be useful to free the memory.
        if free_memory:
            del output
            del tokens
            torch.cuda.synchronize()

        return vecs

    def _get_embedding_fnames(self):
        folder_name = os.path.join(self._config.embeddings.embeddings_path, "splade")
        fn_embeddings = os.path.join(folder_name, "splade_embeddings.npz")
        fn_ids = os.path.join(folder_name, "splade_ids.pickle")
        fn_metadatas = os.path.join(folder_name, "splade_metadatas.pickle")
        return folder_name, fn_embeddings, fn_ids, fn_metadatas

    def load(self) -> None:
        """Loads embeddings from disk."""

        _, fn_embeddings, fn_ids, fn_metadatas = self._get_embedding_fnames()
        try:
            self._embeddings = scipy.sparse.load_npz(fn_embeddings)
            with open(fn_ids, "rb") as fp:
                self._ids = np.array(pickle.load(fp))

            # Load metadata
            with open(fn_metadatas, "rb") as fm:
                self._metadatas = np.array(pickle.load(fm))

            self._l2_norm_matrix = scipy.sparse.linalg.norm(self._embeddings, axis=1)

            for ind, m in enumerate(self._metadatas):
                if m["label"]:
                    self._labels_to_ind[m["label"]].append(ind)

                self._chunk_size_to_ind[m["chunk_size"]].append(ind)

            logger.info(f"SPLADE: Got {len(self._labels_to_ind)} labels.")

        except FileNotFoundError:
            raise FileNotFoundError(
                "Embeddings don't exist, run generate_embeddings_from_docs(..) first."
            )
        logger.info(f"Loaded sparse (SPLADE) embeddings from {fn_embeddings}")

    def generate_embeddings_from_docs(
        self, docs: List[Document], persist: bool = True
    ) -> Tuple[np.ndarray, List[str], List[dict]]:
        """Generates SPLADE embeddings from documents, in batches of chunk_size.

        Args:
            docs (List[Document]): list of Documents to generate the embeddings for.
            chunk_size (int, optional): Number of documents to process per batch.
                                        Defaults to 5. Can be higher for larger VRAM.
        """

        chunk_size = self.n_batch
        logger.info(
            f"Calculating SPLADE embeddings for {len(docs)} documents. Using chunk size: {chunk_size}"
        )

        ids = [d.metadata["document_id"] for d in docs]
        metadatas = [d.metadata for d in docs]

        vecs = []
        for chunk in tqdm.tqdm(
            split(docs, chunk_size=chunk_size), total=int(len(docs) / chunk_size)
        ):
            texts = [d.page_content for d in chunk if d.page_content]
            vecs.append(self._get_batch_embeddings(texts))

        embeddings = np.vstack(vecs)
        logger.info(f"Shape of the embeddings matrix: {embeddings.shape}")

        if persist:
            self.persist_embeddings(embeddings, metadatas, ids)
        return embeddings, ids, metadatas

    def add_embeddings(self, new_docs: List[Document]) -> None:
        if self._embeddings is None or self._ids is None:
            logger.info("Loading embeddings...")
            self.load()

        embeddings, ids, metadatas = self.generate_embeddings_from_docs(
            docs=new_docs, persist=False
        )
        if self._ids is not None and self._embeddings is not None:
            logger.debug(
                f"Splade embeddings shape before update: {self._embeddings.shape}"
            )
            updated_embeddings = np.vstack((self._embeddings.toarray(), embeddings))
            logger.debug(
                f"Splade embeddings shape after update: {updated_embeddings.shape}"
            )
            updated_ids = self._ids.tolist() + ids
            updated_metadata = self._metadatas.tolist() + metadatas
            logger.debug(f"Splade: Updated metadata length: {len(updated_metadata)}")
            logger.debug(f"Splade: Updated ids length: {len(updated_ids)}")
            self.persist_embeddings(updated_embeddings, updated_metadata, updated_ids)
        else:
            raise Exception(
                "Something is wrong: ids and embeddings weren't loaded properly."
            )

    def persist_embeddings(self, embeddings, metadatas, ids):
        folder_name, fn_embeddings, fn_ids, fn_metadatas = self._get_embedding_fnames()
        csr_embeddings = scipy.sparse.csr_matrix(embeddings)

        if not os.path.exists(folder_name):
            os.makedirs(folder_name)

        scipy.sparse.save_npz(fn_embeddings, csr_embeddings)
        self.save_list(ids, fn_ids)
        self.save_list(metadatas, fn_metadatas)
        logger.info(f"Saved SPLADE embeddings to {fn_embeddings}")

    def delete_by_ids(self, delete_ids: List[str]):
        if self._embeddings is None or self._ids is None:
            logger.info("Loading embeddings...")
            self.load()

        if self._ids is not None and self._embeddings is not None:
            indices = [
                ind for ind, id_ in enumerate(self._ids) if id_ not in set(delete_ids)
            ]
            self._ids = self._ids[indices]
            self._embeddings = self._embeddings[indices]
            self._metadatas = self._metadatas[indices]

        else:
            raise Exception(
                "Something is wrong: ids and embeddings weren't loaded properly."
            )

    def query(
        self, search: str, chunk_size: int, n: int = 50, label: str = ""
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Finds documents using sparse index similar to the search query in meaning

        Args:
            search (str): Search query
            n (int, optional): Number of document ids to return. Defaults to 50.

        Raises:
            Exception: If no persitent embeedings were found

        Returns:
            List[str], List[float]: list of document ids with scores, most similar to the search query
        """

        if self._embeddings is None or self._ids is None:
            logger.info("Loading embeddings...")
            self.load()

        # If label is present, restrict the search to a subset of documents containing the label
        if (
            label
            and label in self._labels_to_ind
            and self._embeddings is not None
            and self._ids is not None
        ):
            indices = sorted(
                list(
                    set(self._labels_to_ind[label]).intersection(
                        set(self._chunk_size_to_ind[chunk_size])
                    )
                )
            )
            logger.info(
                f"SPLADE - restricting search to label: {label}, chunk size: {chunk_size}. Number of docs: {len(indices)}"
            )

        else:
            indices = sorted(list(set(self._chunk_size_to_ind[chunk_size])))
            logger.info(
                f"SPLADE search will search over all documents of chunk size: {chunk_size}. Number of docs: {len(indices)}"
            )

        # print(indices)
        embeddings = self._embeddings[indices]  # type: ignore
        ids = self._ids[indices]  # type: ignore
        l2_norm_matrix = scipy.sparse.linalg.norm(embeddings, axis=1)

        embed_query = self._get_batch_embeddings(docs=[search])
        l2_norm_query = scipy.linalg.norm(embed_query)

        if embeddings is not None and l2_norm_matrix is not None and ids is not None:
            cosine_similarity = embeddings.dot(embed_query) / (
                l2_norm_matrix * l2_norm_query
            )
            # print(cosine_similarity)
            most_similar = np.argsort(cosine_similarity)

            top_similar_indices = most_similar[-n:][::-1]
            return (
                ids[top_similar_indices],
                cosine_similarity[top_similar_indices],
            )
        else:
            raise Exception(
                "Something went wrong..Embeddings weren't calculated or loaded properly."
            )

    def save_list(self, list_: list, fname: str) -> None:
        # store list in binary file so 'wb' mode
        with open(fname, "wb") as fp:
            pickle.dump(list_, fp)



import os
from dataclasses import dataclass
from typing import List, Optional, Union

from langchain.chains.base import Chain
from langchain.chains.question_answering import load_qa_chain
from loguru import logger

from llmsearch.chroma import VectorStoreChroma
from llmsearch.splade import SparseEmbeddingsSplade
from llmsearch.config import Config, RerankerModel
from llmsearch.models.utils import get_llm
from llmsearch.ranking import BGEReranker, MarcoReranker
from llmsearch.embeddings import VectorStore
from llmsearch.database.config import DBSettings, get_local_session, Base
from langchain.chains import LLMChain
from langchain.prompts import PromptTemplate

CHAIN_TYPE = "stuff"


@dataclass
class LLMBundle:
    chain: Chain
    store: VectorStore
    reranker: Optional[Union[BGEReranker, MarcoReranker]]
    sparse_search: SparseEmbeddingsSplade
    chunk_sizes: List[int]
    response_persist_db_settings: Optional[DBSettings] = None
    hyde_chain: Optional[LLMChain] = None
    hyde_enabled: bool = False
    multiquery_chain: Optional[LLMChain] = None
    multiquery_enabled: bool = False


def set_cache_folder(cache_folder_root: str):
    """Set temporary cache folder for HF models and transformers"""

    sentence_transformers_home = cache_folder_root
    transformers_cache = os.path.join(cache_folder_root, "transformers")
    hf_home = os.path.join(cache_folder_root, "hf_home")

    logger.info(
        f"Setting SENTENCE_TRANSFORMERS_HOME folder: {sentence_transformers_home}"
    )
    logger.info(f"Setting TRANSFORMERS_CACHE folder: {transformers_cache}")
    logger.info(f"Setting HF_HOME: {hf_home}")
    logger.info(f"Setting MODELS_CACHE_FOLDER: {cache_folder_root}")

    os.environ["SENTENCE_TRANSFORMERS_HOME"] = sentence_transformers_home
    os.environ["TRANSFORMERS_CACHE"] = transformers_cache
    os.environ["HF_HOME"] = hf_home
    os.environ["MODELS_CACHE_FOLDER"] = cache_folder_root


def get_llm_bundle(config: Config) -> LLMBundle:
    """Bundles tools needed for retrieval into auxilarry data structure

    Args:
        config (Config): instance of configuration

    Returns:
        LLMBundle:
    """

    set_cache_folder(str(config.cache_folder))
    llm = get_llm(config.llm.params)  # type: ignore
    chain = load_qa_chain(llm=llm.model, chain_type=CHAIN_TYPE, prompt=llm.prompt)

    store = VectorStoreChroma(
        persist_folder=str(config.embeddings.embeddings_path), config=config
    )
    store._load_retriever()

    reranker = None
    if config.semantic_search.reranker.enabled:
        if config.semantic_search.reranker.model == RerankerModel.BGE_RERANKER:
            reranker = BGEReranker()
        elif config.semantic_search.reranker.model == RerankerModel.MARCO_RERANKER:
            reranker = MarcoReranker()
        else:
            raise TypeError(
                "Invalid reranker type: {}", config.semantic_search.reranker.model
            )

    chunk_sizes = config.embeddings.chunk_sizes

    splade = SparseEmbeddingsSplade(config=config)
    splade.load()

    if config.persist_response_db_path is not None:
        db_settings = get_local_session(db_path=config.persist_response_db_path)
        Base.metadata.create_all(bind=db_settings.engine)
        logger.info("Initialized persistence db.")
    else:
        db_settings = None

    hyde_chain = get_hyde_chain(config, llm.model)
    multiquery_chain = get_multiquery_chain(config, llm.model)

    return LLMBundle(
        chain=chain,
        reranker=reranker,
        chunk_sizes=chunk_sizes,
        sparse_search=splade,
        store=store,
        response_persist_db_settings=db_settings,
        hyde_chain=hyde_chain,
        hyde_enabled=config.semantic_search.hyde.enabled,
        multiquery_chain=multiquery_chain,
        multiquery_enabled=config.semantic_search.multiquery.enabled,
    )


def get_hyde_chain(config, llm_model) -> LLMChain:
    logger.info("Creating HyDE chain...")
    return LLMChain(
        llm=llm_model,
        prompt=PromptTemplate(
            template=config.semantic_search.hyde.hyde_prompt,
            input_variables=["question"],
        ),
    )


def get_multiquery_chain(config, llm_model) -> LLMChain:
    logger.info("Creating MultiQUery chain...")
    return LLMChain(
        llm=llm_model,
        prompt=PromptTemplate(
            template=config.semantic_search.multiquery.multiquery_prompt,
            input_variables=["question", "n_versions"],
        ),
    )



from pydantic import BaseModel, ConfigDict
from pathlib import Path
from typing import Optional, Union


class LlamaModelConfig(BaseModel):
    model_config = ConfigDict()
    model_config["protected_namespaces"] = ()

    model_path: Union[Path, str]
    prompt_template: str
    model_init_params: dict = {}
    model_kwargs: dict = {}


class OpenAIModelConfig(BaseModel):
    model_config = ConfigDict()
    model_config["protected_namespaces"] = ()

    prompt_template: str
    model_kwargs: dict = {}


class AzureOpenAIModelConfig(BaseModel):
    model_config = ConfigDict()
    model_config["protected_namespaces"] = ()

    prompt_template: str
    deployment_name: str
    model_name: str
    model_kwargs: dict = {}
    openai_api_type: str = "azure"
    openai_api_version: str = "2023-05-15"
    openai_api_base: str


# class AutoGPTQModelConfig(BaseModel):
#     model_folder: Union[str, Path]
#     prompt_template: str
#     device: str = "auto"
#     tokenzier_kwargs: dict = {}
#     trust_remote_code: bool = False
#     use_safetensors: bool = True
#     model_kwargs: dict = {}
#     pipeline_kwargs: dict = {}


class HuggingFaceModelConfig(BaseModel):
    model_config = ConfigDict()
    model_config["protected_namespaces"] = ()

    cache_folder: Optional[Union[Path, str]] = None  # will be copied from
    tokenizer_name: Optional[str] = None
    model_name: str
    prompt_template: str
    load_8bit: bool = False
    trust_remote_code: bool = False
    tokenzier_kwargs: dict = {}
    model_kwargs: dict = {}
    pipeline_kwargs: dict = {}
